@@ -17,9 +17,11 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
 
+#include <ctype.h>
+
 #include "common.h"
 #include "cgilib.h"
-#include <ctype.h>
+#include "parseconf.h"
 
 static char *unescape(char *buf)
 {
@@ -146,38 +148,55 @@ void extractpostargs(void)
 	}
 }
 
-int checkhost(char *check, char **desc)
+/* called for fatal errors in parseconf like malloc failures */
+void cgilib_err(const char *errmsg)
 {
-	FILE	*hostlist;
-	char	fn[SMALLBUF], buf[SMALLBUF], addr[SMALLBUF], *rest, *ptr;
-	int	ofs;
+	upslogx(LOG_ERR, "Fatal error in parseconf(ups.conf): %s", errmsg);
+}
 
-	if (!check)
+int checkhost(const char *host, char **desc)
+{
+	char	fn[SMALLBUF];
+	PCONF_CTX	ctx;
+
+	if (!host)
 		return 0;		/* deny null hostnames */
 
-	snprintf(fn, sizeof(fn), "%s/hosts.conf", CONFPATH);
-	hostlist = fopen(fn, "r");
+	snprintf(fn, sizeof(fn), "%s/hosts.conf", confpath());
 
-	if (hostlist == NULL)
-		return 0;		/* default to deny */
+	pconf_init(&ctx, cgilib_err);
 
-	while (fgets(buf, sizeof(buf), hostlist)) 
-		if (strncmp("MONITOR", buf, 7) == 0) {
-			sscanf(buf, "%*s %s %n", addr, &ofs);
-			rest = buf + ofs + 1;
-			ptr = strchr(rest, '"');
-			if (ptr)
-				*ptr  = '\0';
+	if (!pconf_file_begin(&ctx, fn)) {
+		pconf_finish(&ctx);
+		fprintf(stderr, "%s\n", ctx.errmsg);
 
-			if (!strcmp(addr, check)) {
-				fclose(hostlist);
-				if (desc != NULL)
-					*desc = xstrdup(rest);
+		return 0;	/* failed: deny access */
+	}
 
-				return 1;	/* allowed */
-			}
+	while (pconf_file_next(&ctx)) {
+		if (pconf_parse_error(&ctx)) {
+			fprintf(stderr, "Error: %s:%d: %s\n",
+				fn, ctx.linenum, ctx.errmsg);
+			continue;
 		}
 
-	fclose(hostlist);
-	return 0;		/* denied */
+		/* MONITOR <host> <description> */
+		if (ctx.numargs < 3)
+			continue;
+
+		if (strcmp(ctx.arglist[0], "MONITOR") != 0)
+			continue;
+
+		if (!strcmp(ctx.arglist[1], host)) {
+			if (desc)
+				*desc = xstrdup(ctx.arglist[2]);
+
+			pconf_finish(&ctx);
+			return 1;	/* found: allow access */
+		}
+	}
+
+	pconf_finish(&ctx);
+
+	return 0;	/* not found: access denied */
 }	
