@@ -25,6 +25,11 @@
 
 #include "upsclient.h"
 
+struct list_t {
+	char	*name;
+	struct	list_t	*next;
+};
+
 static void usage(char *prog)
 {
 	printf("Network UPS Tools upscmd %s\n\n", UPS_VERSION);
@@ -57,34 +62,24 @@ static void clean_exit(UPSCONN *ups, char *upsname, char *hostname, int code)
 	exit(code);
 }
 
-static void listvars(char *rawname)
+static int old_list(UPSCONN *ups, char *upsname, char *rawname)
 {
 	char	buf[SMALLBUF], buf2[SMALLBUF], *ptr, *tmp, *q, *desc;
 
-	int	port;
-	char	*upsname, *hostname;
-	UPSCONN	ups;
-
-	upscli_splitname(rawname, &upsname, &hostname, &port);
-
-	if (upscli_connect(&ups, hostname, port, 0) < 0) {
-		fprintf(stderr, "Error: %s\n", upscli_strerror(&ups));
-		clean_exit(&ups, upsname, hostname, 1);
-	}
-
-	if (upscli_getlist(&ups, upsname, UPSCLI_LIST_CMDS, buf, 
+	if (upscli_getlist(ups, upsname, UPSCLI_LIST_CMDS, buf, 
 		sizeof(buf)) < 0) {
 
 		fprintf(stderr, "Can't list instant commands: %s\n",
-			upscli_strerror(&ups));
+			upscli_strerror(ups));
 
-		clean_exit(&ups, upsname, hostname, 1);
+		return 1;
 	}
 
 	if (strlen(buf) == 0) {
 		fprintf(stderr, "No instant commands are supported on UPS [%s]\n",
 		       rawname);
-		clean_exit(&ups, upsname, hostname, 1);
+
+		return 1;
 	}
 
 	printf("Instant commands supported on UPS [%s]:\n\n", rawname);
@@ -105,15 +100,15 @@ static void listvars(char *rawname)
 
 		snprintf(buf2, sizeof(buf2), "INSTCMDDESC %s\n", tmp);
 
-		if (upscli_sendline(&ups, buf2, strlen(buf2)) < 0) {
+		if (upscli_sendline(ups, buf2, strlen(buf2)) < 0) {
 			fprintf(stderr, "Description fetch failed: %s\n", 
-				upscli_strerror(&ups));
+				upscli_strerror(ups));
 			continue;
 		}
 
-		if (upscli_readline(&ups, buf2, sizeof(buf2)) < 0) {
+		if (upscli_readline(ups, buf2, sizeof(buf2)) < 0) {
 			fprintf(stderr, "Description fetch failed: %s\n", 
-				upscli_strerror(&ups));
+				upscli_strerror(ups));
 			continue;
 		}
 
@@ -134,13 +129,172 @@ static void listvars(char *rawname)
 		printf("%s\n", desc);
 	}
 
+	return 0;
+}
+
+static void print_cmd(UPSCONN *ups, const char *upsname, char *cmdname)
+{
+	int	ret, numq, numa;
+	char	**answer;
+	const	char	*query[4];
+
+	query[0] = "CMDDESC";
+	query[1] = upsname;
+	query[2] = cmdname;
+	numq = 3;
+
+	ret = upscli_get(ups, numq, query, &numa, &answer);
+
+	if ((ret < 0) || (numa < numq)) {
+		printf("%s\n", cmdname);
+		return;
+	}
+
+	/* CMDDESC <upsname> <cmdname> <desc> */
+	printf("%s - %s\n", cmdname, answer[3]);
+}
+
+static void listcmds(char *rawname)
+{
+	int	ret, port, numq, numa;
+	char	*upsname, *hostname, **answer;
+	const	char	*query[4];
+	UPSCONN	ups;
+	struct	list_t	*lhead, *llast, *ltmp, *lnext;
+
+	upscli_splitname(rawname, &upsname, &hostname, &port);
+
+	if (upscli_connect(&ups, hostname, port, 0) < 0) {
+		fprintf(stderr, "Error: %s\n", upscli_strerror(&ups));
+		clean_exit(&ups, upsname, hostname, 1);
+	}
+
+	if (!upsname) {
+		ret = old_list(&ups, upsname, rawname);
+		clean_exit(&ups, upsname, hostname, ret);
+	}
+
+	llast = lhead = NULL;
+
+	query[0] = "CMD";
+	query[1] = upsname;
+	numq = 2;
+
+	ret = upscli_list_start(&ups, numq, query);
+
+	if (ret < 0) {
+
+		/* old upsd: fall back on old LISTINSTCMD technique */
+		if (upscli_upserror(&ups) == UPSCLI_ERR_UNKCOMMAND) {
+			ret = old_list(&ups, upsname, rawname);
+			clean_exit(&ups, upsname, hostname, ret);
+		}
+
+		fprintf(stderr, "Error: %s\n", upscli_strerror(&ups));
+		clean_exit(&ups, upsname, hostname, 1);
+	}
+
+	printf("Instant commands supported on UPS [%s]:\n\n", rawname);
+
+	ret = upscli_list_next(&ups, numq, query, &numa, &answer);
+
+	while (ret == 1) {
+
+		/* CMD <upsname> <cmdname> */
+
+		if (numa < 3) {
+			fprintf(stderr, "Error: insufficient data "
+				"(got %d args, need at least 3)\n", numa);
+			clean_exit(&ups, upsname, hostname, 1);
+		}
+
+		/* sock this entry away for later */
+
+		ltmp = xmalloc(sizeof(struct list_t));
+		ltmp->name = xstrdup(answer[2]);
+		ltmp->next = NULL;
+
+		if (llast)
+			llast->next = ltmp;
+		else
+			lhead = ltmp;
+
+		llast = ltmp;
+
+		ret = upscli_list_next(&ups, numq, query, &numa, &answer);
+	}
+
+	/* walk the list and try to get descriptions, freeing as we go */
+
+	ltmp = lhead;
+
+	while (ltmp) {
+		lnext = ltmp->next;
+
+		print_cmd(&ups, upsname, ltmp->name);
+
+		free(ltmp->name);
+		free(ltmp);
+		ltmp = lnext;
+	}
+
 	clean_exit(&ups, upsname, hostname, 0);
 }
 
+static int old_do_cmd(UPSCONN *ups, const char *upsname, const char *cmd)
+{
+	char 	buf[SMALLBUF];
+
+	if (upsname)
+		snprintf(buf, sizeof(buf), "INSTCMD %s@%s\n", cmd, upsname);
+	else
+		snprintf(buf, sizeof(buf), "INSTCMD %s\n", cmd);
+
+	if (upscli_sendline(ups, buf, strlen(buf)) < 0) {
+		fprintf(stderr, "Can't send instant command: %s\n",
+			upscli_strerror(ups));
+
+		return 1;
+	}
+
+	if (upscli_readline(ups, buf, sizeof(buf)) < 0) {
+		fprintf(stderr, "Instant command failed: %s\n", 
+			upscli_strerror(ups));
+
+		return 1;
+	}
+
+	return 0;
+}
+
+static int new_do_cmd(UPSCONN *ups, const char *upsname, const char *cmd)
+{
+	char 	buf[SMALLBUF];
+
+	snprintf(buf, sizeof(buf), "INSTCMD %s %s\n", upsname, cmd);
+
+	if (upscli_sendline(ups, buf, strlen(buf)) < 0) {
+		fprintf(stderr, "Can't send instant command: %s\n",
+			upscli_strerror(ups));
+
+		return 1;
+	}
+
+	if (upscli_readline(ups, buf, sizeof(buf)) < 0) {
+		fprintf(stderr, "Instant command failed: %s\n", 
+			upscli_strerror(ups));
+
+		return 1;
+	}
+
+	return 0;
+}
+
 int main(int argc, char **argv)
 {
 	int	have_un, have_pw;
-	char 	*prog, buf[SMALLBUF], username[SMALLBUF], password[SMALLBUF];
+	char 	*prog, buf[SMALLBUF], username[SMALLBUF], password[SMALLBUF],
+		*pwtmp;
 	int	i, ret;
 
 	int	port;
@@ -151,10 +305,10 @@ int main(int argc, char **argv)
 
 	have_un = have_pw = 0;
 
-	while ((i = getopt(argc, argv, "+l:hu:p:")) != EOF) {
+	while ((i = getopt(argc, argv, "+l:hu:p:V")) != EOF) {
 		switch (i) {
 			case 'l':
-				listvars(optarg);
+				listcmds(optarg);
 				break;
 
 			case 'u':
@@ -167,7 +321,12 @@ int main(int argc, char **argv)
 				snprintf(password, sizeof(password), "%s", 
 					optarg);
 				have_pw = 1;
-				break;			
+				break;
+
+			case 'V':
+				printf("Network UPS Tools upscmd %s\n",
+					UPS_VERSION);
+				exit(0);
 
 			case 'h':
 			default:
@@ -200,8 +359,14 @@ int main(int argc, char **argv)
 
 	/* getpass leaks slightly - use -p when testing in valgrind */
 	if (!have_pw) {
-		snprintf(password, sizeof(password), "%s",
-			getpass("Password: "));
+		pwtmp = GETPASS("Password: ");
+
+		if (!pwtmp) {
+			fprintf(stderr, "getpass failed: %s\n", strerror(errno));
+			clean_exit(&ups, upsname, hostname, 1);
+		}
+			
+		snprintf(password, sizeof(password), "%s", pwtmp);
 	}
 
 	snprintf(buf, sizeof(buf), "USERNAME %s\n", username);
@@ -216,7 +381,7 @@ int main(int argc, char **argv)
 	ret = upscli_readline(&ups, buf, sizeof(buf));
 
 	if (ret < 0) {
-		if (ups.upserror == UPSCLI_ERR_UNKCOMMAND) {
+		if (upscli_upserror(&ups) == UPSCLI_ERR_UNKCOMMAND) {
 			fprintf(stderr, "Set username failed due to an "
 				"unknown command.\n");
 
@@ -245,27 +410,20 @@ int main(int argc, char **argv)
 		clean_exit(&ups, upsname, hostname, 1);
 	}
 
-	if (upsname)
-		snprintf(buf, sizeof(buf), "INSTCMD %s@%s\n", argv[1], 
-		         upsname);
-	else
-		snprintf(buf, sizeof(buf), "INSTCMD %s\n", argv[1]);
-
-	if (upscli_sendline(&ups, buf, strlen(buf)) < 0) {
-		fprintf(stderr, "Can't send instant command: %s\n",
-			upscli_strerror(&ups));
-
-		clean_exit(&ups, upsname, hostname, 1);
+	/* no upsname = fallback until 2.0 */
+	if (!upsname) {
+		ret = old_do_cmd(&ups, upsname, argv[1]);
+		clean_exit(&ups, upsname, hostname, ret);
 	}
 
-	if (upscli_readline(&ups, buf, sizeof(buf)) < 0) {
-		fprintf(stderr, "Instant command failed: %s\n", 
-			upscli_strerror(&ups));
-
-		clean_exit(&ups, upsname, hostname, 1);
+	/* also fallback for old command names */
+	if (!strchr(argv[1], '.')) {
+		ret = old_do_cmd(&ups, upsname, argv[1]);
+		clean_exit(&ups, upsname, hostname, ret);
 	}
 
-	clean_exit(&ups, upsname, hostname, 0);
+	ret = new_do_cmd(&ups, upsname, argv[1]);
+	clean_exit(&ups, upsname, hostname, ret);
 
 	/* NOTREACHED */
 	return 0;
