@@ -1,5 +1,12 @@
 /* upslog.h - table of functions for handling various logging functions */
 
+/* function list */
+struct flist_t {
+	void	(*fptr)(const char *arg);
+	const	char	*arg;
+	struct	flist_t	*next;
+};
+
 static void do_host(const char *arg);
 static void do_upshost(const char *arg);
 static void do_pid(const char *arg);
