@@ -59,10 +59,13 @@ static	int	nocommwarntime = 300;
 static	int	finaldelay = 5;
 
 	/* set by SIGHUP handler, cleared after reload finishes */
-static	int	reload = 0;
+static	int	reload_flag = 0;
+
+	/* set after SIGINT, SIGQUIT, or SIGTERM */
+static	int	exit_flag = 0;
 
 	/* userid for unprivileged process when using fork mode */
-static	char	*run_as_user;
+static	char	*run_as_user = NULL;
 
 	/* SSL details - where to find certs, whether to use them */
 static	char	*certpath = NULL;
@@ -135,6 +138,9 @@ static void notify(const char *notice, int flags, const char *ntype,
 	char	exec[LARGEBUF];
 	int	ret;
 
+	if (flag_isset(flags, NOTIFY_IGNORE))
+		return;
+
 	if (flag_isset(flags, NOTIFY_SYSLOG))
 		upslogx(LOG_NOTICE, "%s", notice);
 
@@ -182,11 +188,12 @@ static void do_notify(const utype *ups, int ntype)
 
 	for (i = 0; notifylist[i].name != NULL; i++) {
 		if (notifylist[i].type == ntype) {
-			debug("do_notify: ntype 0x%04x (%s)\n", ntype, notifylist[i].name);
-
-			snprintf(msg, sizeof(msg), notifylist[i].msg, ups ? ups->sys : "");
-
-			notify(msg, notifylist[i].flags, notifylist[i].name, upsname);
+			debug("do_notify: ntype 0x%04x (%s)\n", ntype, 
+				notifylist[i].name);
+			snprintf(msg, sizeof(msg), notifylist[i].msg, 
+				ups ? ups->sys : "");
+			notify(msg, notifylist[i].flags, notifylist[i].name, 
+				upsname);
 			return;
 		}
 	}
@@ -195,28 +202,31 @@ static void do_notify(const utype *ups, int ntype)
 }
 
 /* check for master permissions on the server for this ups */
-static void checkmaster(utype *ups)
+static int checkmaster(utype *ups)
 {
 	char	buf[SMALLBUF];
 
 	/* don't bother if we're not configured as a master for this ups */
 	if (!flag_isset(ups->status, ST_MASTER))
-		return;
+		return 1;
 
-	if (ups->upsname)
-		snprintf(buf, sizeof(buf), "MASTER %s\n", ups->upsname);
-	else
-		snprintf(buf, sizeof(buf), "MASTER\n");
+	/* this shouldn't happen (LOGIN checks it earlier) */
+	if ((ups->upsname == NULL) || (strlen(ups->upsname) == 0)) {
+		upslogx(LOG_ERR, "Set master on UPS [%s] failed: empty upsname");
+		return 0;
+	}
+
+	snprintf(buf, sizeof(buf), "MASTER %s\n", ups->upsname);
 
 	if (upscli_sendline(&ups->conn, buf, strlen(buf)) < 0) {
 		upslogx(LOG_ALERT, "Can't set master mode on UPS [%s] - %s",
 			ups->sys, upscli_strerror(&ups->conn));
-		return;
+		return 0;
 	}
 
 	if (upscli_readline(&ups->conn, buf, sizeof(buf)) == 0) {
 		if (!strncmp(buf, "OK", 2))
-			return;
+			return 1;
 
 		/* not ERR, but not caught by readline either? */
 
@@ -229,30 +239,34 @@ static void checkmaster(utype *ups)
 			ups->sys);
 		upslogx(LOG_ALERT, "Reason: %s", upscli_strerror(&ups->conn));
 	}
+
+	return 0;
 }
 
-/* authenticate to upsd and login */
-static void do_login(utype *ups)
+/* authenticate to upsd, plus do LOGIN and MASTER if applicable */
+static int do_upsd_auth(utype *ups)
 {
 	char	buf[SMALLBUF];
 
-	/* use it if we have it, but don't require it yet */
+	if (!ups->un) {
+		upslogx(LOG_ERR, "UPS [%s]: no username defined!", ups->sys);
+		return 0;
+	}
 
-	if (ups->un) {
-		snprintf(buf, sizeof(buf), "USERNAME %s\n", ups->un);
-		if (upscli_sendline(&ups->conn, buf, strlen(buf)) < 0) {
-			upslogx(LOG_ERR, "Can't set username on [%s]: %s",
-				ups->sys, upscli_strerror(&ups->conn));
-				return;
-		}
+	if (ups->pv == 0)	/* monitor only, no need to login */
+		return 1;
 
-		if (upscli_readline(&ups->conn, buf, sizeof(buf)) < 0) {
-			upslogx(LOG_ERR, "Set username on [%s] failed (%s) -"
-				" assuming ancient upsd",
-				ups->sys, upscli_strerror(&ups->conn));
+	snprintf(buf, sizeof(buf), "USERNAME %s\n", ups->un);
+	if (upscli_sendline(&ups->conn, buf, strlen(buf)) < 0) {
+		upslogx(LOG_ERR, "Can't set username on [%s]: %s",
+			ups->sys, upscli_strerror(&ups->conn));
+			return 0;
+	}
 
-			/* just a warning for now - it's not required yet */
-		}
+	if (upscli_readline(&ups->conn, buf, sizeof(buf)) < 0) {
+		upslogx(LOG_ERR, "Set username on [%s] failed: %s",
+			ups->sys, upscli_strerror(&ups->conn));
+		return 0;
 	}
 
 	/* authenticate first */
@@ -261,46 +275,48 @@ static void do_login(utype *ups)
 	if (upscli_sendline(&ups->conn, buf, strlen(buf)) < 0) {
 		upslogx(LOG_ERR, "Can't set password on [%s]: %s",
 			ups->sys, upscli_strerror(&ups->conn));
-			return;
+			return 0;
 	}
 
 	if (upscli_readline(&ups->conn, buf, sizeof(buf)) < 0) {
 		upslogx(LOG_ERR, "Set password on [%s] failed: %s",
 			ups->sys, upscli_strerror(&ups->conn));
-		return;
+		return 0;
 	}
 
 	/* catch insanity from the server - not ERR and not OK either */
 	if (strncmp(buf, "OK", 2) != 0) {
 		upslogx(LOG_ERR, "Set password on [%s] failed - got [%s]",
 			ups->sys, buf);
-		return;
+		return 0;
+	}
+
+	/* we require a upsname now */
+	if ((ups->upsname == NULL) || (strlen(ups->upsname) == 0)) {
+		upslogx(LOG_ERR, "Login to UPS [%s] failed: empty upsname");
+		return 0;
 	}
 
 	/* password is set, let's login */
-
-	if ((ups->upsname == NULL) || (strlen(ups->upsname) == 0))
-		snprintf(buf, sizeof(buf), "LOGIN\n");
-	else
-		snprintf(buf, sizeof(buf), "LOGIN %s\n", ups->upsname);
+	snprintf(buf, sizeof(buf), "LOGIN %s\n", ups->upsname);
 
 	if (upscli_sendline(&ups->conn, buf, strlen(buf)) < 0) {
 		upslogx(LOG_ERR, "Login to UPS [%s] failed: %s",
 			ups->sys, upscli_strerror(&ups->conn));
-		return;
+		return 0;
 	}
 
 	if (upscli_readline(&ups->conn, buf, sizeof(buf)) < 0) {
 		upslogx(LOG_ERR, "Can't login to UPS [%s]: %s",
 			ups->sys, upscli_strerror(&ups->conn));
-		return;
+		return 0;
 	}
 
 	/* catch insanity from the server - not ERR and not OK either */
 	if (strncmp(buf, "OK", 2) != 0) {
 		upslogx(LOG_ERR, "Login on UPS [%s] failed - got [%s]",
 			ups->sys, buf);
-		return;
+		return 0;
 	}
 
 	/* finally - everything is OK */
@@ -308,69 +324,69 @@ static void do_login(utype *ups)
 	setflag(&ups->status, ST_LOGIN);
 
 	/* now see if we also need to test master permissions */
-	checkmaster(ups);
+	return checkmaster(ups);
 }
 
 /* set flags and make announcements when a UPS has been checked successfully */
-static void heardups(utype *ups)
+static void ups_is_alive(utype *ups)
 {
 	time_t	now;
 
 	time(&now);
 	ups->lastpoll = now;
 
-	if (flag_isset(ups->status, ST_ALIVE))
+	if (ups->commstate == 1)		/* already known */
 		return;
 
-	setflag(&ups->status, ST_ALIVE);
-
-	/* only notify on subsequent reconnects */
-	if (!flag_isset(ups->status, ST_FIRST))
-		setflag(&ups->status, ST_FIRST);
-	else
+	/* only notify for 0->1 transitions (to ignore the first connect) */
+	if (ups->commstate == 0)
 		do_notify(ups, NOTIFY_COMMOK);
 
-	if (ups->pv == 0)	/* monitor only, no need to login */
-		return;
-
-	/* if not already logged in, do it now */
-	if (!flag_isset(ups->status, ST_LOGIN))
-		do_login(ups);
+	ups->commstate = 1;
 }
 
-static void upsgone(utype *ups)
+/* handle all the notifications for a missing UPS in one place */
+static void ups_is_gone(utype *ups)
 {
-	/* if the UPS is already known to be gone, then do nothing here */
-	if (!flag_isset(ups->status, ST_ALIVE))
-		return;
+	time_t	now;
 
-	/* don't clear OL/OB/LB flags since we may use them later */
-	clearflag(&ups->status, ST_ALIVE);
-	clearflag(&ups->status, ST_LOGIN);
+	/* first time: clear the flag and throw the first notifier */
+	if (ups->commstate != 0) {
+		ups->commstate = 0;
 
-	do_notify(ups, NOTIFY_COMMBAD);
-}
-
-static void upsonbatt(utype *ups)
-{
-	sleepval = pollfreqalert;	/* bump up polling frequency */
-
-	if (!flag_isset(ups->status, ST_ALIVE)) {
-		debug("upsonbatt(%s) (first time)\n", ups->sys);
-		heardups(ups);
-		do_notify(ups, NOTIFY_ONBATT);
-		setflag(&ups->status, ST_ONBATT);
-		clearflag(&ups->status, ST_ONLINE);
+		/* COMMBAD is the initial loss of communications */
+		do_notify(ups, NOTIFY_COMMBAD);
 		return;
 	}
 
-	debug("upsonbatt(%s) (normal)\n", ups->sys);
-	
-	heardups(ups);
+	time(&now);
 
-	if (flag_isset(ups->status, ST_ONBATT)) 	/* no change */
+	/* first only act if we're <nocommtime> seconds past the last poll */
+	if ((now - ups->lastpoll) < nocommwarntime)
 		return;
 
+	/* now only complain if we haven't lately */
+	if ((now - ups->lastncwarn) > nocommwarntime) {
+
+		/* NOCOMM indicates a persistent condition */
+		do_notify(ups, NOTIFY_NOCOMM);
+		ups->lastncwarn = now;
+	}
+}
+
+static void ups_on_batt(utype *ups)
+{
+	sleepval = pollfreqalert;	/* bump up polling frequency */
+
+	if (flag_isset(ups->status, ST_ONBATT)) { 	/* no change */
+		debug("ups_on_batt(%s) (no change)\n", ups->sys);
+		return;
+	}
+
+	ups->linestate = 0;	
+
+	debug("ups_on_batt(%s) (first time)\n", ups->sys);
+
 	/* must have changed from OL to OB, so notify */
 
 	do_notify(ups, NOTIFY_ONBATT);
@@ -378,26 +394,21 @@ static void upsonbatt(utype *ups)
 	clearflag(&ups->status, ST_ONLINE);
 }
 
-static void upsonline(utype *ups)
+static void ups_on_line(utype *ups)
 {
-	if (!flag_isset(ups->status, ST_ALIVE)) {
-		debug("upsonline(%s) (first time)\n", ups->sys);
-		heardups(ups);
-		setflag(&ups->status, ST_ONLINE);
-		clearflag(&ups->status, ST_ONBATT);
+	if (flag_isset(ups->status, ST_ONLINE)) { 	/* no change */
+		debug("ups_on_line(%s) (no change)\n", ups->sys);
 		return;
 	}
 
-	debug("upsonline(%s) (normal)\n", ups->sys);
+	debug("ups_on_line(%s) (first time)\n", ups->sys);
 
-	heardups(ups);
+	/* ignore the first OL at startup, otherwise send the notifier */
+	if (ups->linestate != -1)
+		do_notify(ups, NOTIFY_ONLINE);
 
-	if (flag_isset(ups->status, ST_ONLINE)) 	/* no change */
-		return;
+	ups->linestate = 1;
 
-	/* must have changed from OB to OL, so notify */
-
-	do_notify(ups, NOTIFY_ONLINE);
 	setflag(&ups->status, ST_ONLINE);
 	clearflag(&ups->status, ST_ONBATT);
 }
@@ -495,6 +506,109 @@ static void setfsd(utype *ups)
 		        buf);
 }
 
+static void set_alarm(void)
+{
+	alarm(NET_TIMEOUT);
+}
+
+static void clear_alarm(void)
+{
+	signal(SIGALRM, SIG_IGN);
+	alarm(0);
+}
+
+static int old_get_var(utype *ups, const char *var, char *buf, size_t bufsize)
+{
+	debug("old_get_var: %s / %s\n", ups->sys, var);
+
+	if (!strcmp(var, "numlogins"))
+		return upscli_getvar(&ups->conn, ups->upsname, "numlogins",
+ 			buf, bufsize);
+
+	if (!strcmp(var, "status"))
+		return upscli_getvar(&ups->conn, ups->upsname, "status",
+			buf, bufsize);
+
+	upslogx(LOG_ERR, "old_get_var: programming error: var=%s", var);
+	return -1;
+}
+
+static int new_get_var(utype *ups, const char *var, char *buf, size_t bufsize)
+{
+	int	ret, numq, numa;
+	const	char	*query[4];
+	char	**answer;
+
+	/* can't do the new stuff without a UPS name set */
+
+	/* this shouldn't happen (we now require upsnames) */
+	if (!ups->upsname) {
+		upslogx(LOG_INFO, "UPS [%s]: no upsname set, forcing compatibility mode",
+			ups->sys);
+		ups->old = 1;
+		return old_get_var(ups, var, buf, bufsize);
+	}
+
+	numq = 0;
+
+	if (!strcmp(var, "numlogins")) {
+		query[0] = "NUMLOGINS";
+		query[1] = ups->upsname;
+		numq = 2;
+	}
+
+	if (!strcmp(var, "status")) {
+		query[0] = "VAR";
+		query[1] = ups->upsname;
+		query[2] = "ups.status";
+		numq = 3;
+	}
+
+	if (numq == 0) {
+		upslogx(LOG_ERR, "new_get_var: programming error: var=%s", var);
+		return -1;
+	}
+
+	debug("new_get_var: %s / %s\n", ups->sys, var);
+
+	ret = upscli_get(&ups->conn, numq, query, &numa, &answer);
+
+	if (ret < 0) {
+
+		/* detect old upsd */
+		if (upscli_upserror(&ups->conn) == UPSCLI_ERR_UNKCOMMAND) {
+
+			upslogx(LOG_INFO, "UPS [%s]: using compatibility mode",
+				ups->sys);
+
+			ups->old = 1;
+			return old_get_var(ups, var, buf, bufsize);
+		}
+
+		/* some other error */
+		return -1;
+	}
+
+	if (numa < numq) {
+		upslogx(LOG_ERR, "%s: Error: insufficient data "
+			"(got %d args, need at least %d)\n", 
+			var, numa, numq);
+		return -1;
+	}
+
+	snprintf(buf, bufsize, "%s", answer[numq]);
+	return 0;
+}
+
+static int get_var(utype *ups, const char *var, char *buf, size_t bufsize)
+{
+	if (ups->old)
+		return old_get_var(ups, var, buf, bufsize);
+
+	/* try it the new way */
+	return new_get_var(ups, var, buf, bufsize);
+}
+
 static void slavesync(void)
 {
 	utype	*ups;
@@ -513,13 +627,16 @@ static void slavesync(void)
 			if (!flag_isset(ups->status, ST_MASTER))
 				continue;
 
-			if (upscli_getvar(&ups->conn, ups->upsname, "numlogins", 
-			                temp, sizeof(temp)) >= 0) {
+			set_alarm();
+
+			if (get_var(ups, "numlogins", temp, sizeof(temp)) >= 0) {
 				logins = strtol(temp, (char **)NULL, 10);
 
 				if (logins > maxlogins)
 					maxlogins = logins;
 			}
+
+			clear_alarm();
 		}
 
 		/* if no UPS has more than 1 login (us), then slaves are gone */
@@ -640,18 +757,14 @@ static void recalc(void)
 		forceshutdown();
 }		
 
-static void upslowbatt(utype *ups)
+static void ups_low_batt(utype *ups)
 {
-	if (!flag_isset(ups->status, ST_ALIVE)) {
-		heardups(ups);
-		setflag(&ups->status, ST_LOWBATT);
+	if (flag_isset(ups->status, ST_LOWBATT)) { 	/* no change */
+		debug("ups_low_batt(%s) (no change)\n", ups->sys);
 		return;
 	}
 
-	heardups(ups);
-
-	if (flag_isset(ups->status, ST_LOWBATT)) 	/* no change */
-		return;
+	debug("ups_low_batt(%s) (first time)\n", ups->sys);
 
 	/* must have changed from !LB to LB, so notify */
 
@@ -671,36 +784,14 @@ static void upsreplbatt(utype *ups)
 	}
 }
 
-/* deal with a UPS that stays unavailable for awhile */
-static void upsnocomm(utype *ups)
+static void ups_fsd(utype *ups)
 {
-	time_t	now;
-
-	time(&now);
-
-	/* first only act if we're <nocommtime> seconds past the last poll */
-	if ((now - ups->lastpoll) < nocommwarntime)
-		return;
-
-	/* now only complain if we haven't lately */
-	if ((now - ups->lastncwarn) > nocommwarntime) {
-		do_notify(ups, NOTIFY_NOCOMM);
-		ups->lastncwarn = now;
-	}
-}
-
-static void upsfsd(utype *ups)
-{
-	if (!flag_isset(ups->status, ST_ALIVE)) {
-		heardups(ups);
-		setflag(&ups->status, ST_FSD);
+	if (flag_isset(ups->status, ST_FSD)) {		/* no change */
+		debug("ups_fsd(%s) (no change)\n", ups->sys);
 		return;
 	}
 
-	heardups(ups);
-
-	if (flag_isset(ups->status, ST_FSD)) 	/* no change */
-		return;
+	debug("ups_fsd(%s) (first time)\n", ups->sys);
 
 	/* must have changed from !FSD to FSD, so notify */
 
@@ -713,7 +804,11 @@ static void drop_connection(utype *ups)
 {
 	debug("Dropping connection to UPS [%s]\n", ups->sys);
 
-	clearflag(&ups->status, ST_ALIVE);
+	/* reset this in case upsd gets upgraded */
+	ups->old = 0;
+
+	ups->commstate = 0;
+	ups->linestate = 0;
 	clearflag(&ups->status, ST_LOGIN);
 	clearflag(&ups->status, ST_CONNECTED);
 
@@ -829,11 +924,20 @@ static void addups(int reloading, const char *sys, const char *pvs,
 	int	pv;
 	utype	*tmp, *last;
 
-	/* un (username) may be NULL for now - old style host-based auth */
+	/* the username is now required - no more host-based auth */
 
-	if ((!sys) || (!pvs) || (!pw) || (!master)) {
+	if ((!sys) || (!pvs) || (!pw) || (!master) || (!un)) {
 		upslogx(LOG_WARNING, "Ignoring invalid MONITOR line in upsmon.conf!");
-		upslogx(LOG_WARNING, "MONITOR configuration directives require four arguments.");
+		upslogx(LOG_WARNING, "MONITOR configuration directives require five arguments.");
+		return;
+	}
+
+	/* deal with sys without a upsname - refuse it */
+	if (!strchr(sys, '@')) {
+		upslogx(LOG_WARNING, "Ignoring invalid MONITOR line in ups.conf (%s)",
+			sys);
+		upslogx(LOG_WARNING, "UPS directives now require a UPS name "
+			"(MONITOR upsname@hostname ...)");
 		return;
 	}
 
@@ -878,10 +982,16 @@ static void addups(int reloading, const char *sys, const char *pvs,
 	tmp->pw = xstrdup(pw);
 	tmp->status = 0;
 	tmp->retain = 1;
+
+	/* ignore initial COMMOK and ONLINE by default */
+	tmp->commstate = -1;
+	tmp->linestate = -1;
+
 	tmp->lastpoll = 0;
 	tmp->lastnoncrit = 0;
 	tmp->lastrbwarn = 0;
 	tmp->lastncwarn = 0;
+	tmp->old = 0;
 
 	if (!strcasecmp(master, "master"))
 		setflag(&tmp->status, ST_MASTER);
@@ -901,6 +1011,10 @@ static void addups(int reloading, const char *sys, const char *pvs,
 		upslogx(LOG_INFO, "UPS: %s (monitoring only)", tmp->sys);
 
         upscli_splitname(tmp->sys, &tmp->upsname, &tmp->hostname, &tmp->port);
+
+	if (!tmp->upsname)
+		upslogx(LOG_WARNING, "Warning: UPS [%s]: no upsname set!",
+			tmp->sys);
 }		
 
 static void set_notifymsg(const char *name, const char *msg)
@@ -966,7 +1080,7 @@ static void set_notifyflag(const char *ntype, char *flags)
 			newflag = NOTIFY_IGNORE;
 
 		if (newflag)
-			notifylist[i].flags |= newflag;
+			notifylist[pos].flags |= newflag;
 		else
 			upslogx(LOG_WARNING, "Invalid notify flag: [%s]\n", ptr);
 
@@ -998,7 +1112,7 @@ static void checkmode(char *cfgentry, char *oldvalue, char *newvalue,
 	upslogx(LOG_WARNING, "You must restart upsmon for this change to work");
 }
 
-/* returns 1 if used, 0 if not, so conf_arg can complain about bogus configs */
+/* returns 1 if used, 0 if not, so we can complain about bogus configs */
 static int parse_conf_arg(int numargs, char **arg)
 {
 	/* using up to arg[1] below */
@@ -1007,7 +1121,7 @@ static int parse_conf_arg(int numargs, char **arg)
 
 	/* SHUTDOWNCMD <cmd> */
 	if (!strcmp(arg[0], "SHUTDOWNCMD")) {
-		checkmode(arg[0], shutdowncmd, arg[1], reload);
+		checkmode(arg[0], shutdowncmd, arg[1], reload_flag);
 
 		if (shutdowncmd)
 			free(shutdowncmd);
@@ -1017,14 +1131,14 @@ static int parse_conf_arg(int numargs, char **arg)
 
 	/* POWERDOWNFLAG <fn> */
 	if (!strcmp(arg[0], "POWERDOWNFLAG")) {
-		checkmode(arg[0], powerdownflag, arg[1], reload);
+		checkmode(arg[0], powerdownflag, arg[1], reload_flag);
 
 		if (powerdownflag)
 			free(powerdownflag);
 
 		powerdownflag = xstrdup(arg[1]);
 
-		if (!reload)
+		if (!reload_flag)
 			upslogx(LOG_INFO, "Using power down flag file %s\n",
 				arg[1]);
 
@@ -1138,76 +1252,78 @@ static int parse_conf_arg(int numargs, char **arg)
 	if (numargs < 5)
 		return 0;
 
-
 	if (!strcmp(arg[0], "MONITOR")) {
 
 		/* original style: no username (only 5 args) */
-
 		if (numargs == 5) {
+			upslogx(LOG_ERR, "Unable to use old-style MONITOR line without a username");
+			upslogx(LOG_ERR, "Convert it and add a username to upsd.users - see the documentation");
 
-			/* <sys> <pwrval> <user> <pw> ("master" | "slave") */
-			addups(reload, arg[1], arg[2], NULL, arg[3], arg[4]);
-			return 1;
+			fatalx("Fatal error: unusable configuration");
 		}
 
-		/* new style with username: 6 args */
-
-		if (numargs == 6) {
-
-			/* <sys> <pwrval> <user> <pw> ("master" | "slave") */
-			addups(reload, arg[1], arg[2], arg[3], arg[4], arg[5]);
-			return 1;
-		}
+		/* <sys> <pwrval> <user> <pw> ("master" | "slave") */
+		addups(reload_flag, arg[1], arg[2], arg[3], arg[4], arg[5]);
+		return 1;
 	}
 
 	/* didn't parse it at all */
 	return 0;
 }
 
-/* wrap around the actual parser to detect things that aren't handled */
-void conf_arg(int numargs, char **arg)
+/* called for fatal errors in parseconf like malloc failures */
+void upsmon_err(const char *errmsg)
 {
-	if (numargs < 1)
-		return;
-
-	if (!parse_conf_arg(numargs, arg)) {
-		int	i;
-		char	errmsg[SMALLBUF];
-
-		snprintf(errmsg, sizeof(errmsg), "upsmon.conf: invalid directive");
-
-		for (i = 0; i < numargs; i++)
-			snprintfcat(errmsg, sizeof(errmsg), " %s", arg[i]);
-
-		upslogx(LOG_WARNING, errmsg);
-	}
-}
-
-void conf_err(int linenum, char *errtext)
-{
-	if (linenum == 0)
-		upslogx(LOG_ERR, "upsmon.conf: %s", errtext);
-	else
-		upslogx(LOG_ERR, "upsmon.conf line %d: %s", linenum, errtext);
+	upslogx(LOG_ERR, "Fatal error in parseconf(upsmon.conf): %s", errmsg);
 }
 
 static void loadconfig(void)
 {
-	int	ret;
-	char	cfn[SMALLBUF];
+	char	fn[SMALLBUF];
+	PCONF_CTX	ctx;
 
-	snprintf(cfn, sizeof(cfn), "%s/upsmon.conf", CONFPATH);
+	snprintf(fn, sizeof(fn), "%s/upsmon.conf", confpath());
 
-	ret = parseconf(cfn, conf_arg, conf_err);
+	pconf_init(&ctx, upsmon_err);
 
-	if (ret != 0) {
-		if (reload == 1) {
-			upslog(LOG_ERR, "Reload failed: can't open %s", cfn);
+	if (!pconf_file_begin(&ctx, fn)) {
+		pconf_finish(&ctx);
+
+		if (reload_flag == 1) {
+			upslog(LOG_ERR, "Reload failed: %s", ctx.errmsg);
 			return;
 		}
 
-		fatal("Can't parse %s", cfn);
+		fatal("%s", ctx.errmsg);
 	}
+
+	while (pconf_file_next(&ctx)) {
+		if (pconf_parse_error(&ctx)) {
+			upslogx(LOG_ERR, "Parse error: %s:%d: %s",
+				fn, ctx.linenum, ctx.errmsg);
+			continue;
+		}
+
+		if (ctx.numargs < 1)
+			continue;
+
+		if (!parse_conf_arg(ctx.numargs, ctx.arglist)) {
+			int	i;
+			char	errmsg[SMALLBUF];
+
+			snprintf(errmsg, sizeof(errmsg), 
+				"upsmon.conf line %d: invalid directive",
+				ctx.linenum);
+
+			for (i = 0; i < ctx.numargs; i++)
+				snprintfcat(errmsg, sizeof(errmsg), " %s", 
+					ctx.arglist[i]);
+
+			upslogx(LOG_WARNING, errmsg);
+		}
+	}
+
+	pconf_finish(&ctx);		
 }
 
 /* SIGPIPE handler */
@@ -1217,33 +1333,79 @@ static void sigpipe(int sig)
 }
 
 /* SIGQUIT, SIGTERM handler */
-static void shut_myself_down(int sig)
+static void set_exit_flag(int sig)
 {
-	utype	*ups;
-
 	upslogx(LOG_INFO, "Signal %d: exiting", sig);
+	exit_flag = 1;
+}
+
+static void ups_free(utype *ups)
+{
+	if (ups->sys)
+		free(ups->sys);
+	if (ups->upsname)
+		free(ups->upsname);
+	if (ups->hostname)
+		free(ups->hostname);
+	if (ups->un)
+		free(ups->un);
+	if (ups->pw)
+		free(ups->pw);
+	free(ups);
+}
+
+static void upsmon_exit(void)
+{
+	int	i;
+	utype	*utmp, *unext;
 
 	/* close all fds */
-	for (ups = firstups; ups != NULL; ups = ups->next)
-		upscli_disconnect(&ups->conn);
+	utmp = firstups;
+
+	while (utmp) {
+		unext = utmp->next;
+
+		upscli_disconnect(&utmp->conn);
+		ups_free(utmp);
+
+		utmp = unext;
+	}
+
+	if (run_as_user)
+		free(run_as_user);
+	if (shutdowncmd)
+		free(shutdowncmd);
+	if (notifycmd)
+		free(notifycmd);
+	if (powerdownflag)
+		free(powerdownflag);
+
+	for (i = 0; notifylist[i].name != NULL; i++)
+		if (notifylist[i].msg != notifylist[i].stockmsg)
+			free(notifylist[i].msg);
 
 	exit(0);
 }
 
 static void user_fsd(int sig)
 {
-	upslogx(LOG_INFO, "User requested FSD!");
-
+	upslogx(LOG_INFO, "Signal %d: User requested FSD", sig);
 	userfsd = 1;
 }
 
 static void set_reload_flag(int sig)
 {
-	reload = 1;
+	reload_flag = 1;
+}
+
+/* handler for alarm when getupsvarfd times out */
+static void read_timeout(int sig)
+{
+	/* don't do anything here, just return */
 }
 
 /* install handlers for a few signals */
-static void setupsignals(void)
+static void setup_signals(void)
 {
 	sigemptyset(&upsm_sigmask);
 	sa.sa_mask = upsm_sigmask;
@@ -1252,11 +1414,16 @@ static void setupsignals(void)
 	sa.sa_handler = sigpipe;
 	sigaction(SIGPIPE, &sa, NULL);
 
-	sa.sa_handler = shut_myself_down;
+	sa.sa_handler = set_exit_flag;
 	sigaction(SIGINT, &sa, NULL);
 	sigaction(SIGQUIT, &sa, NULL);
 	sigaction(SIGTERM, &sa, NULL);
 
+	/* handle timeouts */
+
+	sa.sa_handler = read_timeout;
+	sigaction(SIGALRM, &sa, NULL);
+
 	/* deal with the ones from userspace as well */
 
 	sa.sa_handler = user_fsd;
@@ -1266,12 +1433,6 @@ static void setupsignals(void)
 	sigaction(SIGCMD_RELOAD, &sa, NULL);
 }
 
-/* handler for alarm when getupsvarfd times out */
-static void timeout(int sig)
-{
-	/* don't do anything here, just return */
-}
-
 /* remember the last time the ups was not critical (OB + LB) */
 static void update_crittimer(utype *ups)
 {
@@ -1286,6 +1447,51 @@ static void update_crittimer(utype *ups)
 	/* fallthrough: let the timer age */
 }
 
+static int try_ssl(utype *ups)
+{
+	int	ret;
+
+	/* if not doing SSL, we're done */
+	if (!upscli_ssl(&ups->conn))
+		return 1;
+
+	if (!certpath) {
+		if (certverify == 1) {
+			upslogx(LOG_ERR, "Configuration error: "
+				"CERTVERIFY is set, but CERTPATH isn't");
+			upslogx(LOG_ERR, "UPS [%s]: Connection impossible, "
+				"dropping link", ups->sys);
+
+			ups_is_gone(ups);
+			drop_connection(ups);
+
+			return 0;	/* failed */
+		}
+
+		/* certverify is 0, so just warn them and return */
+		upslogx(LOG_WARNING, "Certificate verification is disabled");
+		return 1;
+	}
+
+	/* you REALLY should set CERTVERIFY to 1 if using SSL... */
+	if (certverify == 0)
+		upslogx(LOG_WARNING, "Certificate verification is disabled");
+
+	ret = upscli_sslcert(&ups->conn, NULL, certpath, certverify);
+
+	if (ret < 0) {
+		upslogx(LOG_ERR, "UPS [%s]: SSL certificate set failed: %s",
+			ups->sys, upscli_strerror(&ups->conn));
+
+		ups_is_gone(ups);
+		drop_connection(ups);
+
+		return 0;
+	}
+
+	return 1;
+}
+
 /* handle connecting to upsd, plus get SSL going too if possible */
 static int try_connect(utype *ups)
 {
@@ -1306,145 +1512,122 @@ static int try_connect(utype *ups)
 	if (ret < 0) {
 		upslogx(LOG_ERR, "UPS [%s]: connect failed: %s",
 			ups->sys, upscli_strerror(&ups->conn));
-			upsgone(ups);
-		upsnocomm(ups);
+
+		ups_is_gone(ups);
 		return 0;
 	}
 
+	ret = try_ssl(ups);
+
+	if (ret == 0)
+		return 0;	/* something broke while trying SSL */
+
+	/* we're definitely connected now */
 	setflag(&ups->status, ST_CONNECTED);
 
-	/* if not doing SSL, we're done */
-	if (!ups->conn.ssl)
-		return 1;
+	/* now try to authenticate to upsd */
 
-	if (!certpath) {
-		if (certverify == 1) {
-			upslogx(LOG_ERR, "Configuration error: "
-				"CERTVERIFY is set, but CERTPATH isn't");
-			upslogx(LOG_ERR, "UPS [%s]: Connection impossible, "
-				"dropping link", ups->sys);
-			drop_connection(ups);
-			upsgone(ups);
-			upsnocomm(ups);
-			return 0;	/* failed */
-		}
+	ret = do_upsd_auth(ups);
 
-		/* certverify is 0, so just warn them and return */
-		upslogx(LOG_WARNING, "Certificate verification is disabled");
-		return 1;
+	if (ret == 1)
+		return 1;		/* everything is happy */
+
+	/* something failed in the auth so we may not be completely logged in */
+
+	/* FUTURE: do something beyond the error msgs from do_upsd_auth? */
+
+	return 0;
+}
+
+/* deal with the contents of STATUS or ups.status for this ups */
+static void parse_status(utype *ups, char *status)
+{
+	char	*stat, *ptr;
+
+	clear_alarm();
+
+	debug("     status: [%s]\n", status);
+
+	/* empty response is the same as a dead ups */
+	if (!strcmp(status, "")) {
+		ups_is_gone(ups);
+		return;
 	}
 
-	/* you REALLY should set CERTVERIFY to 1 if using SSL... */
-	if (certverify == 0)
-		upslogx(LOG_WARNING, "Certificate verification is disabled");
+	ups_is_alive(ups);
 
-	ret = upscli_sslcert(&ups->conn, NULL, certpath, certverify);
+	/* clear these out early if they disappear */
+	if (!strstr(status, "LB"))
+		clearflag(&ups->status, ST_LOWBATT);
+	if (!strstr(status, "FSD"))
+		clearflag(&ups->status, ST_FSD);
 
-	if (ret < 0) {
-		upslogx(LOG_ERR, "UPS [%s]: SSL certificate set failed: %s",
-			ups->sys, upscli_strerror(&ups->conn));
-		drop_connection(ups);
-		upsgone(ups);
-		upsnocomm(ups);
-		return 0;
-	}
+	stat = status;
 
-	setflag(&ups->status, ST_CONNECTED);
+	/* split up the status words and parse each one separately */
+	while (stat != NULL) {
+		ptr = strchr(stat, ' ');
+		if (ptr)
+			*ptr++ = '\0';
 
-	return 1;
+		debug("    parsing: [%s]: ", stat);
+
+		if (!strcasecmp(stat, "OL"))
+			ups_on_line(ups);
+		if (!strcasecmp(stat, "OB"))
+			ups_on_batt(ups);
+		if (!strcasecmp(stat, "LB"))
+			ups_low_batt(ups);
+		if (!strcasecmp(stat, "RB"))
+			upsreplbatt(ups);
+
+		/* do it last to override any possible OL */
+		if (!strcasecmp(stat, "FSD"))
+			ups_fsd(ups);
+
+		update_crittimer(ups);
+
+		stat = ptr;
+	} 
+
+	debug("\n");
 }
 
 /* see what the status of the UPS is and handle any changes */
 static void pollups(utype *ups)
 {
-	char	status[SMALLBUF], *stat, *ptr;
-	struct	sigaction sa;
-	sigset_t upsm_sigmask;
+	char	status[SMALLBUF];
 
 	/* try a reconnect here */
 	if (!flag_isset(ups->status, ST_CONNECTED))
 		if (try_connect(ups) != 1)
 			return;
 
-	if (ups->conn.ssl)
+	if (upscli_ssl(&ups->conn) == 1)
 		debug("polling ups: %s [SSL]\n", ups->sys);
 	else
 		debug("polling ups: %s\n", ups->sys);
 
-	sa.sa_handler = timeout;
-	sigemptyset(&upsm_sigmask);
-	sa.sa_mask = upsm_sigmask;
-	sa.sa_flags = 0;
-	sigaction(SIGALRM, &sa, NULL);
+	set_alarm();
 
-	alarm(10);
-
-	if (upscli_getvar(&ups->conn, ups->upsname, "status", status, 
-		sizeof(status)) == 0) {
-
-		signal(SIGALRM, SIG_IGN);
-		alarm(0);
-
-		debug("     status: [%s]\n", status);
-
-		/* empty response is the same as a dead ups */
-		if (!strcmp(status, "")) {
-			upsgone(ups);
-			upsnocomm(ups);
-			return;
-		}
-
-		/* clear these out early if they disappear */
-		if (!strstr(status, "LB"))
-			clearflag(&ups->status, ST_LOWBATT);
-		if (!strstr(status, "FSD"))
-			clearflag(&ups->status, ST_FSD);
-
-		stat = status;
-
-		while (stat != NULL) {
-			ptr = strchr(stat, ' ');
-			if (ptr)
-				*ptr++ = '\0';
-
-			debug("    parsing: [%s]: ", stat);
-
-			if (!strcasecmp(stat, "OL"))
-				upsonline(ups);
-			if (!strcasecmp(stat, "OB"))
-				upsonbatt(ups);
-			if (!strcasecmp(stat, "LB"))
-				upslowbatt(ups);
-			if (!strcasecmp(stat, "RB"))
-				upsreplbatt(ups);
-
-			/* do it last to override any possible OL */
-			if (!strcasecmp(stat, "FSD"))
-				upsfsd(ups);
-
-			update_crittimer(ups);
-
-			stat = ptr;
-		} 
-
-		debug("\n");
+	if (get_var(ups, "status", status, sizeof(status)) == 0) {
+		clear_alarm();
+		parse_status(ups, status);
 		return;
 	}
 
 	/* fallthrough: no communications */
-	signal(SIGALRM, SIG_IGN);
-	alarm(0);
+	clear_alarm();
 
 	/* try to make some of these a little friendlier */
-	switch(ups->conn.upserror) {
+
+	switch (upscli_upserror(&ups->conn)) {
+
 		case UPSCLI_ERR_UNKNOWNUPS:
-			if (!ups->upsname)
-				upslogx(LOG_ERR, "Poll UPS [%s] failed - no "
-				"UPS defined on server %s", ups->hostname);
-			else
-				upslogx(LOG_ERR, "Poll UPS [%s] failed - [%s] "
-				"does not exist on server %s", 
-				ups->sys, ups->upsname,	ups->hostname);
+			upslogx(LOG_ERR, "Poll UPS [%s] failed - [%s] "
+			"does not exist on server %s", 
+			ups->sys, ups->upsname,	ups->hostname);
+
 			break;
 		default:
 			upslogx(LOG_ERR, "Poll UPS [%s] failed - %s", 
@@ -1452,21 +1635,14 @@ static void pollups(utype *ups)
 			break;
 	}
 
-	/* restart the connection if we had ever polled it */
-	if (flag_isset(ups->status, ST_ALIVE)) {
+	/* throw COMMBAD or NOCOMM as conditions may warrant */
+	ups_is_gone(ups);
+
+	/* if upsclient lost the connection, clean up things on our side */
+	if (upscli_fd(&ups->conn) == -1) {
 		drop_connection(ups);
-		upsgone(ups);
-		upsnocomm(ups);
 		return;
 	}
-
-	/* if anything worse than staleness/not defined, restart the link */
-	if ((ups->conn.upserror != UPSCLI_ERR_DATASTALE) &&
-		(ups->conn.upserror != UPSCLI_ERR_UNKNOWNUPS))
-		drop_connection(ups);
-
-	upsgone(ups);
-	upsnocomm(ups);
 }
 
 /* remove the power down flag if it exists and is the proper form */
@@ -1504,15 +1680,16 @@ static void clearpdf(void)
 
 static void help(const char *progname)
 {
-	printf("Monitor UPS servers and possibly shutdown if necessary.\n\n");
-	printf("usage: %s [-h] [-d] [-i] [-p] [-u <user>] [-c <cmd>]\n\n", progname);
-	printf("\n");	
+	printf("Monitors UPS servers and may initiate shutdown if necessary.\n\n");
+
+	printf("usage: %s [OPTIONS]\n\n", progname);
 	printf("  -c <cmd>	send command to running process\n");	
 	printf("		commands:\n");
 	printf("		 - fsd: shutdown all master UPSes (use with caution)\n");
 	printf("		 - reload: reread configuration\n");
 	printf("		 - stop: stop monitoring and exit\n");
-	printf("  -d		enable debugging code\n");
+	printf("  -D		raise debugging level\n");
+	printf("  -h		display this help\n");
 	printf("  -i		go into an infinite loop after calling shutdown\n");
 	printf("  -p		always run privileged (disable privileged parent)\n");
 	printf("  -u <user>	run child as user <user> (ignored when using -p)\n");
@@ -1569,6 +1746,15 @@ static void runparent(int pipe)
 static void start_pipe(char *user)
 {
 	int	ret;
+	struct	passwd	*new_uid = NULL;
+
+	/* default to the --with-user value from configure */
+	if (!user)
+		user = RUN_AS_USER;
+
+	/* grab user data early, while we still have the console */
+	if ((new_uid = get_user_pwent(user)) == NULL)
+		fatal("getpwnam(%s)", user);
 
 	ret = pipe(pipefd);
 
@@ -1593,11 +1779,7 @@ static void start_pipe(char *user)
 	/* write the pid file now, as we will soon lose root */
 	writepid("upsmon");
 
-	/* continuing from here is the unprivileged child, so let's drop root */
-	if (user)
-		switch_to_user(user, NULL);
-	else
-		droproot();
+	become_user(new_uid);
 }
 
 static void delete_ups(utype *target)
@@ -1625,16 +1807,7 @@ static void delete_ups(utype *target)
 
 			/* release memory */
 
-			if (ptr->sys)
-				free(ptr->sys);
-
-			if (ptr->upsname)
-				free(ptr->upsname);
-			if (ptr->hostname)
-				free(ptr->hostname);
-			if (ptr->pw)
-				free(ptr->pw);
-			free(ptr);
+			ups_free(ptr);
 
 			return;
 		}
@@ -1653,7 +1826,7 @@ static int check_file(char *fn)
 	char	chkfn[SMALLBUF];
 	FILE	*f;
 
-	snprintf(chkfn, sizeof(chkfn), "%s/%s", CONFPATH, fn);
+	snprintf(chkfn, sizeof(chkfn), "%s/%s", confpath(), fn);
 
 	f = fopen(chkfn, "r");
 
@@ -1674,7 +1847,7 @@ static void reload_conf(void)
 
 	/* sanity check */
 	if (!check_file("upsmon.conf")) {
-		reload = 0;
+		reload_flag = 0;
 		return;
 	}
 
@@ -1714,7 +1887,7 @@ static void reload_conf(void)
 	}
 
 	/* finally clear the flag */
-	reload = 0;
+	reload_flag = 0;
 }
 
 /* make sure the parent is still alive */
@@ -1760,10 +1933,11 @@ int main(int argc, char *argv[])
 
 	printf("Network UPS Tools upsmon %s\n", UPS_VERSION);
 
-	while ((i = getopt(argc, argv, "+dhic:pu:")) != EOF) {
+	while ((i = getopt(argc, argv, "+Ddhic:pu:V")) != EOF) {
 		switch (i) {
+			case 'D':
 			case 'd':
-				debuglevel = 1;
+				debuglevel++;
 				break;
 			case 'h':
 				help(argv[0]);
@@ -1787,8 +1961,11 @@ int main(int argc, char *argv[])
 					help(argv[0]);
 				break;
 			case 'u':
-				run_as_user = optarg;
+				run_as_user = xstrdup(optarg);
 				break;
+			case 'V':
+				/* just show the banner */
+				exit(0);
 			default:
 				help(argv[0]);
 				break;
@@ -1834,12 +2011,12 @@ int main(int argc, char *argv[])
 	if (use_pipe)
 		start_pipe(run_as_user);
 	else {
-		upslog(LOG_INFO, "Not using split process model.");
+		upslogx(LOG_INFO, "Warning: running as one big root process by request (upsmon -p)");
 		writepid("upsmon");
 	}
 
-	/* ignore sigpipes before the connection opens */
-	setupsignals();
+	/* prep our signal handlers */
+	setup_signals();
 
 	/* reopen the log for the child process */
 	closelog();
@@ -1852,7 +2029,10 @@ int main(int argc, char *argv[])
 		if (userfsd)
 			forceshutdown();
 
-		if (reload)
+		if (exit_flag)
+			upsmon_exit();
+
+		if (reload_flag)
 			reload_conf();
 
 		sleepval = pollfreq;
