@@ -24,9 +24,9 @@
 #define ST_LOWBATT      0x004  	/* UPS has a low battery (LB)		*/
 #define ST_FSD		0x008	/* master has set forced shutdown flag	*/
 #define ST_MASTER       0x010	/* we are the master on this UPS	*/
-#define ST_ALIVE        0x020	/* UPS is connected and responding	*/
+/* was ST_ALIVE 0x020 */
 #define ST_LOGIN        0x040  	/* we are logged into this UPS		*/
-#define ST_FIRST	0x080	/* set after first poll finishes	*/
+/* was ST_FIRST 0x080 */
 #define ST_CONNECTED	0x100	/* upscli_connect returned OK		*/
 
 /* required contents of flag file */
@@ -42,11 +42,18 @@ typedef struct {
 	char	*hostname;		/* just hostname		*/
 	int	port;			/* just the port		*/
 
+	int	old;			/* use REQ instead of GET, etc.	*/
+
 	int     pv;			/* power value from conf	*/
 	char	*un;			/* username (optional for now)	*/
 	char    *pw;  			/* password from conf		*/
 	int     status;			/* status (see flags above)	*/
 	int	retain;			/* tracks deletions at reload	*/
+
+	/* handle suppression of COMMOK and ONLINE at startup */
+	int	commstate;		/* these start at -1, and only	*/
+	int	linestate;		/* fire on a 0->1 transition	*/
+
 	time_t	lastpoll;		/* time of last successful poll	*/
 	time_t  lastnoncrit;		/* time of last non-crit poll	*/
 	time_t	lastrbwarn;		/* time of last REPLBATT warning*/
@@ -103,3 +110,7 @@ struct {
 #define SIGCMD_FSD	SIGUSR1
 #define SIGCMD_STOP	SIGTERM
 #define SIGCMD_RELOAD	SIGHUP
+
+/* various constants */
+
+#define NET_TIMEOUT 10		/* wait 10 seconds max for upsd to respond */
