@@ -20,6 +20,7 @@
 #include "common.h"
 
 #include <netdb.h>
+#include <stdlib.h>
 #include <netinet/in.h>
 #include <sys/socket.h>
 
@@ -27,6 +28,11 @@
 #include "cgilib.h"
 #include "parseconf.h"
 
+struct list_t {
+	char	*name;
+	struct	list_t	*next;
+};
+
 /* see the stock upsset.conf for the whole rant on what this is */
 #define MAGIC_ENABLE_STRING "I_HAVE_SECURED_MY_CGI_DIRECTORY"
 
@@ -135,7 +141,8 @@ static void do_enum(const char *varname)
 {
 	char	out[SMALLBUF], temp[SMALLBUF], *val, *sel, *ptr;
 
-	snprintf(out, sizeof(out), "ENUM %s\n", varname);	
+	snprintf(out, sizeof(out),
+		(upsname ? "ENUM %s@%s\n" : "ENUM %s\n"), varname, upsname);
 	
 	if (upscli_sendline(&ups, out, strlen(out)) < 0) {
 		fprintf(stderr, "ENUM %s failed: %s\n", varname,
@@ -263,7 +270,7 @@ static void do_hidden(const char *next)
 }
 
 /* generate SELECT chooser from hosts.conf entries */
-void upslist_arg(int numargs, char **arg)
+static void upslist_arg(int numargs, char **arg)
 {
 	if (numargs < 3)
 		return;
@@ -280,37 +287,55 @@ void upslist_arg(int numargs, char **arg)
 	}
 }
 
-void upslist_err(int linenum, char *errtext)
+/* called for fatal errors in parseconf like malloc failures */
+void upsset_hosts_err(const char *errmsg)
 {
-	if (linenum == 0)
-		fprintf(stderr, "hosts.conf: %s\n", errtext);
-	else
-		fprintf(stderr, "hosts.conf line %d: %s\n", linenum, errtext);
+	upslogx(LOG_ERR, "Fatal error in parseconf(hosts.conf): %s", errmsg);
 }
 
 /* this defaults to wherever we are now, ups and function-wise */
 static void do_pickups(const char *currfunc)
 {
 	char	hostfn[SMALLBUF];
-	int	ret;
+	PCONF_CTX	ctx;
 
-	snprintf(hostfn, sizeof(hostfn), "%s/hosts.conf", CONFPATH);
+	snprintf(hostfn, sizeof(hostfn), "%s/hosts.conf", confpath());
 
 	printf("<FORM METHOD=\"POST\" ACTION=\"upsset.cgi\">\n");
 
 	printf("Select UPS and function:\n<BR>\n");
-	printf("<SELECT NAME=\"monups\">\n");
 
-	ret = parseconf(hostfn, upslist_arg, upslist_err);
+	pconf_init(&ctx, upsset_hosts_err);
 
-	printf("</SELECT>\n");
+	if (!pconf_file_begin(&ctx, hostfn)) {
+		pconf_finish(&ctx);
 
-	if (ret != 0) {
 		printf("Error: hosts.conf unavailable\n");
 		printf("</FORM>\n");
+
+		/* stderr is for the admin - should wind up in error.log */
+		fprintf(stderr, "upsset: %s\n", ctx.errmsg);
+
 		return;
 	}
 
+	printf("<SELECT NAME=\"monups\">\n");
+
+	while (pconf_file_next(&ctx)) {
+		if (pconf_parse_error(&ctx)) {
+			upslogx(LOG_ERR, "Parse error: %s:%d: %s",
+				hostfn, ctx.linenum, ctx.errmsg);
+                        
+			continue;
+		}
+
+		upslist_arg(ctx.numargs, ctx.arglist);		
+	}
+
+	pconf_finish(&ctx);
+
+	printf("</SELECT>\n");
+
 	printf("<SELECT NAME=\"function\">\n");
 
 	/* FUTURE */
@@ -335,21 +360,32 @@ static void do_pickups(const char *currfunc)
 	printf("</FORM>\n");
 }
 
-static void errorpage(const char *title, const char *msg, const char *func)
+static void error_page(const char *next, const char *title,
+	const char *fmt, ...)
 {
+	char	msg[SMALLBUF];
+	va_list	ap;
+
+	va_start(ap, fmt);
+	vsnprintf(msg, sizeof(msg), fmt, ap);
+	va_end(ap);
+
 	do_header(title);
+
 	start_table();
 	printf("<TR><TH COLSPAN=2 BGCOLOR=\"#60B0B0\">\n");
 	printf("Error: %s\n", msg);
 	printf("</TH></TR>\n");
 
 	printf("<TR><TD ALIGN=\"CENTER\" COLSPAN=2>\n");
-	do_pickups(func);
+	do_pickups(next);
 	printf("</TD></TR>\n");
 
 	printf("</TABLE>\n");
 	printf("</TD></TR></TABLE>\n");
 	printf("</BODY></HTML>\n");
+
+	upscli_disconnect(&ups);
 	exit(0);
 }
 
@@ -377,6 +413,8 @@ static void loginscreen(void)
 	printf("</FORM>\n");
 	printf("</TD></TR></TABLE>\n");
 	printf("</BODY></HTML>\n");
+
+	upscli_disconnect(&ups);
 	exit(0);
 }
 
@@ -386,42 +424,30 @@ static void upsd_connect(void)
 	upscli_splitname(monups, &upsname, &hostname, &port);
 
 	if (upscli_connect(&ups, hostname, port, 0) < 0) {
-		char	ebuf[SMALLBUF];
-
-		snprintf(ebuf, sizeof(ebuf), 
-		         "Unable to connect to %s - %s", hostname,
-		         upscli_strerror(&ups));
-		errorpage("Connect failure", ebuf, "showsettings");
+		error_page("showsettings", "Connect failure",
+			"Unable to connect to %s: %s",
+			monups, upscli_strerror(&ups));
+		/* NOTREACHED */
 	}
 }
 
-/* generate a list of instant commands */
-static void showcmds(void)
+static void old_showcmds(const char *desc)
 {
-	char	*desc, *ptr, cmds[SMALLBUF], *tmp;
-
-	if (!checkhost(monups, &desc))
-		errorpage("Access denied",
-		          "Access to that host is not authorized",
-		          "showsettings");
-
-	upsd_connect();
+	char	*ptr, cmds[SMALLBUF], *tmp;
 
 	if (upscli_getlist(&ups, upsname, UPSCLI_LIST_CMDS, cmds, 
 		sizeof(cmds)) < 0) {
 
-		char	ebuf[SMALLBUF];
-
-		snprintf(ebuf, sizeof(ebuf),
+		error_page("showcmds", "Protocol failure",
 		         "Unable to get list of commands - %s\n",
 		         upscli_strerror(&ups));
-		errorpage("Protocol failure", ebuf, "showcmds");
+
+		/* NOTREACHED */
 	}
 
 	if (strlen(cmds) == 0)
-		errorpage("No instant commands supported",
-		          "This UPS doesn't support any instant commands.",
-		          "showcmds");
+		error_page("showcmds", "No instant commands supported",
+		          "This UPS doesn't support any instant commands.");
 
 	do_header("Instant commands");
 	printf("<FORM ACTION=\"upsset.cgi\" METHOD=\"POST\">\n");
@@ -506,69 +532,174 @@ static void showcmds(void)
 	printf("</TABLE>\n");
 	printf("</BODY></HTML>\n");
 
+	upscli_disconnect(&ups);
 	exit(0);
 }
 
-/* take a command from the client and send it to upsd */
-static void docmd(void)
+static void print_cmd(const char *cmd)
 {
-	char	*desc, buf[SMALLBUF];
+	int	ret, numq, numa;
+	char	**answer;
+	const	char	*query[4];
+
+	query[0] = "CMDDESC";
+	query[1] = upsname;
+	query[2] = cmd;
+	numq = 3;
+
+	ret = upscli_get(&ups, numq, query, &numa, &answer);
+
+	if ((ret < 0) || (numa < numq))
+		return;
+
+	/* CMDDESC <upsname> <cmdname> <desc> */
+
+	printf("<OPTION VALUE=\"%s\">%s</OPTION>\n", cmd, answer[3]);
+}
+
+static void new_showcmds(const char *desc)
+{
+	int	ret, numq, numa;
+	const	char	*query[2];
+	char	**answer;
+	struct	list_t	*lhead, *llast, *ltmp, *lnext;
+
+	llast = lhead = NULL;
+
+	query[0] = "CMD";
+	query[1] = upsname;
+	numq = 2;
+
+	ret = upscli_list_start(&ups, numq, query);
+
+	if (ret < 0) {
+
+		/* old upsd: fall back */
+
+		if (upscli_upserror(&ups) == UPSCLI_ERR_UNKCOMMAND)
+			old_showcmds(desc);
+
+		fprintf(stderr, "LIST CMD %s failed: %s\n",
+			upsname, upscli_strerror(&ups));
+
+		error_page("showcmds", "Server protocol error",
+			"LIST CMD command failed");
+
+		/* NOTREACHED */
+	}
+
+	ret = upscli_list_next(&ups, numq, query, &numa, &answer);
+
+	while (ret == 1) {
+
+		/* CMD upsname cmdname */
+		if (numa < 3) {
+			fprintf(stderr, "Error: insufficient data "
+				"(got %d args, need at least 3)\n", numa);
+
+			return;
+		}
+
+		ltmp = xmalloc(sizeof(struct list_t));
+		ltmp->name = xstrdup(answer[2]);
+		ltmp->next = NULL;
+
+		if (llast)
+			llast->next = ltmp;
+		else
+			lhead = ltmp;
+
+		llast = ltmp;
+
+		ret = upscli_list_next(&ups, numq, query, &numa, &answer);
+	}
+
+	if (!lhead)
+		error_page("showcmds", "No instant commands supported",
+		          "This UPS doesn't support any instant commands.");
+
+	do_header("Instant commands");
+	printf("<FORM ACTION=\"upsset.cgi\" METHOD=\"POST\">\n");
+	start_table();
+
+	/* include the description from checkhost() if present */
+	if (desc)
+		printf("<TR><TH BGCOLOR=\"#60B0B0\"COLSPAN=2>%s</TH></TR>\n",
+		       desc);
+
+	printf("<TR BGCOLOR=\"#60B0B0\" ALIGN=\"CENTER\">\n");
+	printf("<TD>Instant commands</TD>\n");
+
+	printf("<TD>\n");
+	printf("<SELECT NAME=\"upscommand\">\n");
+
+	/* provide a dummy do-nothing default choice */
+	printf("<OPTION VALUE=\"\" SELECTED></OPTION>\n");
+
+	ltmp = lhead;
+
+	while (ltmp) {
+		lnext = ltmp->next;
+
+		print_cmd(ltmp->name);
+
+		free(ltmp->name);
+		free(ltmp);
+		ltmp = lnext;
+	}
+
+	printf("</SELECT>\n");
+	printf("</TD></TR>\n");
+
+	printf("<TR BGCOLOR=\"#60B0B0\">\n");
+	printf("<TD COLSPAN=\"2\" ALIGN=\"CENTER\">\n");
+	do_hidden("docmd");
+	printf("<INPUT TYPE=\"HIDDEN\" NAME=\"monups\" VALUE=\"%s\">\n", monups);
+	printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"Issue command\">\n");
+	printf("<INPUT TYPE=\"RESET\" VALUE=\"Reset\">\n");
+	printf("</TD></TR>\n");
+	printf("</TABLE>\n");
+	printf("</FORM>\n");
+
+	printf("<TR><TD ALIGN=\"CENTER\">\n");
+	do_pickups("showcmds");
+	printf("</TD></TR>\n");
+
+	printf("</TABLE>\n");
+	printf("</BODY></HTML>\n");
+
+	upscli_disconnect(&ups);
+	exit(0);
+}	
+
+/* generate a list of instant commands */
+static void showcmds(void)
+{
+	char	*desc;
 
 	if (!checkhost(monups, &desc))
-		errorpage("Access denied",
-		          "Access to that host is not authorized",
-		          "showsettings");
-
-	/* the user is messing with us */
-	if (!upscommand)	
-		errorpage("Form error", "No instant command selected", "showcmds");
-
-	/* (l)user took the default blank option */
-	if (strlen(upscommand) == 0)
-		errorpage("Form error", "No instant command selected", "showcmds");
+		error_page("showsettings", "Access denied",
+		          "Access to that host is not authorized");
 
 	upsd_connect();
 
-	/* log into server - it makes the authentication decision now */
-	snprintf(buf, sizeof(buf), "USERNAME %s\n", username);
+	if (!upsname)
+		old_showcmds(desc);
+	else
+		new_showcmds(desc);
 
-	if (upscli_sendline(&ups, buf, strlen(buf)) < 0) {
-		printf("Can't set username\n");
+	/* NOTREACHED */
+}
 
-		fprintf(stderr, "Can't set username: %s\n",
-			upscli_strerror(&ups));
-		return;
-	}
-
-	if (upscli_readline(&ups, buf, sizeof(buf)) < 0) {
-
-		/* test for old upsd that doesn't do USERNAME */
-		if (ups.upserror == UPSCLI_ERR_UNKCOMMAND) {
-			errorpage("Protocol mismatch",
-			          "upsd version too old - USERNAME not supported",
-				  "showcmds");
-		}
-
-		errorpage("Can't set username", upscli_strerror(&ups),
-			   "showcmds");
-	}
-
-	snprintf(buf, sizeof(buf), "PASSWORD %s\n", password);
-
-	if (upscli_sendline(&ups, buf, strlen(buf)) < 0)
-		errorpage("Password set failed", upscli_strerror(&ups),
-		          "showcmds");
-
-	if (upscli_readline(&ups, buf, sizeof(buf)) < 0)
-		errorpage("Password set failed", upscli_strerror(&ups),
-		          "showcmds");
+static void old_docmd(const char *desc, const char *cmd)
+{
+	char	buf[SMALLBUF];
 
 	/* send command, get response, show it */
 	if (upsname)
-		snprintf(buf, sizeof(buf), "INSTCMD %s@%s\n", upscommand,
-		         upsname);
+		snprintf(buf, sizeof(buf), "INSTCMD %s@%s\n", cmd, upsname);
 	else
-		snprintf(buf, sizeof(buf), "INSTCMD %s\n", upscommand);
+		snprintf(buf, sizeof(buf), "INSTCMD %s\n", cmd);
 
 	if (upscli_sendline(&ups, buf, strlen(buf)) < 0) {
 		do_header("Error while issuing command");
@@ -586,6 +717,7 @@ static void docmd(void)
 
 		printf("</TD></TR></TABLE>\n");
 		printf("</BODY></HTML>\n");
+		upscli_disconnect(&ups);
 		exit(0);
 	}
 
@@ -605,6 +737,7 @@ static void docmd(void)
 
 		printf("</TD></TR></TABLE>\n");
 		printf("</BODY></HTML>\n");
+		upscli_disconnect(&ups);
 		exit(0);
 	}
 
@@ -624,44 +757,177 @@ static void docmd(void)
 	printf("</TD></TR></TABLE>\n");
 	printf("</BODY></HTML>\n");
 
+	upscli_disconnect(&ups);
+	exit(0);
+
+}
+
+static void new_docmd(const char *desc, const char *cmd)
+{
+	char	buf[SMALLBUF];
+
+	snprintf(buf, sizeof(buf), "INSTCMD %s %s\n", upsname, cmd);
+
+	if (upscli_sendline(&ups, buf, strlen(buf)) < 0) {
+		do_header("Error while issuing command");
+
+		start_table();    
+
+		printf("<TR><TD>Error sending command: %s\n</TD></TR>",
+			upscli_strerror(&ups));
+
+		printf("<TR><TD ALIGN=\"CENTER\" COLSPAN=2>\n");
+		do_pickups("showcmds");
+		printf("</TD></TR>\n");
+
+		printf("</TABLE>\n");
+
+		printf("</TD></TR></TABLE>\n");
+		printf("</BODY></HTML>\n");
+		upscli_disconnect(&ups);
+		exit(0);
+	}
+
+	if (upscli_readline(&ups, buf, sizeof(buf)) < 0) {
+		do_header("Error while reading command response");
+
+		start_table();    
+
+		printf("<TR><TD>Error reading command response: %s\n</TD></TR>",
+			upscli_strerror(&ups));
+
+		printf("<TR><TD ALIGN=\"CENTER\" COLSPAN=2>\n");
+		do_pickups("showcmds");
+		printf("</TD></TR>\n");
+
+		printf("</TABLE>\n");
+
+		printf("</TD></TR></TABLE>\n");
+		printf("</BODY></HTML>\n");
+		upscli_disconnect(&ups);
+		exit(0);
+	}
+
+	do_header("Issuing command");
+	start_table();
+
+	printf("<TR><TD><PRE>\n");
+	printf("Sending command: %s\n", upscommand);
+	printf("Response: %s\n", buf);
+	printf("</PRE></TD></TR>\n");
+
+	printf("<TR><TD ALIGN=\"CENTER\" COLSPAN=2>\n");
+	do_pickups("showcmds");
+	printf("</TD></TR>\n");
+
+	printf("</TABLE>\n");
+	printf("</TD></TR></TABLE>\n");
+	printf("</BODY></HTML>\n");
+
+	upscli_disconnect(&ups);
 	exit(0);
 }
 
-/* dump a table of all read/write variables and their current values */
-static void showsettings(void)
+/* handle setting authentication data in the server */
+static void send_auth(const char *next)
 {
-	char	vars[SMALLBUF], temp[SMALLBUF], out[SMALLBUF], *v, 
-		*type, *ptr, *desc, *vptr;
-	int	typelen;
+	char	buf[SMALLBUF];
+
+	snprintf(buf, sizeof(buf), "USERNAME %s\n", username);
+
+	if (upscli_sendline(&ups, buf, strlen(buf)) < 0) {
+		fprintf(stderr, "Can't set username: %s\n",
+			upscli_strerror(&ups));
+
+		error_page(next, "Can't set username",
+			"Set username failed: %s", upscli_strerror(&ups));
+	}
+
+	if (upscli_readline(&ups, buf, sizeof(buf)) < 0) {
+
+		/* test for old upsd that doesn't do USERNAME */
+		if (upscli_upserror(&ups) == UPSCLI_ERR_UNKCOMMAND) {
+			error_page(next, "Protocol mismatch",
+			          "upsd version too old - USERNAME not supported");
+		}
+
+		error_page(next, "Can't set user name", 
+			"Set user name failed: %s", upscli_strerror(&ups));
+	}
+
+	snprintf(buf, sizeof(buf), "PASSWORD %s\n", password);
+
+	if (upscli_sendline(&ups, buf, strlen(buf)) < 0)
+		error_page(next, "Can't set password",
+			"Password set failed: %s", upscli_strerror(&ups));
+
+	if (upscli_readline(&ups, buf, sizeof(buf)) < 0)
+		error_page(next, "Can't set password",
+			"Password set failed: %s", upscli_strerror(&ups));
+}	
+
+/* take a command from the client and send it to upsd */
+static void docmd(void)
+{
+	char	*desc;
 
 	if (!checkhost(monups, &desc))
-		errorpage("Access denied",
-		          "Access to that host is not authorized",
-		          "showsettings");
+		error_page("showsettings", "Access denied",
+		          "Access to that host is not authorized");
+
+	/* the user is messing with us */
+	if (!upscommand)	
+		error_page("showcmds", "Form error", 
+			"No instant command selected");
+
+	/* (l)user took the default blank option */
+	if (strlen(upscommand) == 0)
+		error_page("showcmds", "Form error", 
+			"No instant command selected");
 
 	upsd_connect();
 
+	send_auth("showcmds");
+
+	if (!upsname)
+		old_docmd(desc, upscommand);
+
+	if (!strchr(upscommand, '.'))
+		old_docmd(desc, upscommand);
+
+	new_docmd(desc, upscommand);
+
+	/* NOTREACHED */
+}
+
+/* dump a table of all read/write variables and their current values */
+static void old_showsettings(const char *desc)
+{
+	char	vars[SMALLBUF], temp[SMALLBUF], out[SMALLBUF], *v, 
+		*type, *ptr, *vptr;
+	int	typelen;
+
 	if (upscli_getlist(&ups, upsname, UPSCLI_LIST_RW, vars, 
 		sizeof(vars)) < 0) {
 
-		char	ebuf[SMALLBUF];
-		snprintf(ebuf, sizeof(ebuf), 
+		error_page("showsettings", "Server protocol error",
 		         "Unable to get variable list - %s\n",
 		         upscli_strerror(&ups));
-		errorpage("Server protocol error", ebuf, "showsettings");
+
+		/* NOTREACHED */
 	}
 
 	if (strlen(vars) == 0) {
-		errorpage("No read/write variables", 
-		          "This UPS has no read/write variables.",
-		          "showsettings");
+		error_page("showsettings", "No read/write variables", 
+		          "This UPS has no read/write variables.");
+		/* NOTREACHED */
 	}
 
 	/* no spaces = seriously broken string */
 	if (!strchr(vars, ' ')) {		
-		errorpage("Server protocol error", 
-		          "Unable to get variable list - broken string", 
-		          "showsettings");
+		error_page("showsettings", "Server protocol error", 
+		          "Unable to get variable list - broken string");
+		/* NOTREACHED */
 	}
 
 	do_header("Current settings");
@@ -723,7 +989,8 @@ static void showsettings(void)
 		printf("<TD>%s</TD>\n", ptr);
 
 		/* now get variable type */
-		snprintf(out, sizeof(out), "VARTYPE %s\n", v);
+		snprintf(out, sizeof(out),
+			(upsname ? "VARTYPE %s@%s\n" : "VARTYPE %s\n"), v, upsname);
 
 		if (upscli_sendline(&ups, out, strlen(out)) < 0) {
 			printf("Can't get VARTYPE\n");
@@ -789,59 +1056,377 @@ static void showsettings(void)
 	printf("</TABLE>\n");
 	printf("</BODY></HTML>\n");
 
+	upscli_disconnect(&ups);
 	exit(0);
 }
 
+static const char *get_data(const char *type, const char *varname)
+{
+	int	ret, numq, numa;
+	char	**answer;
+	const	char	*query[4];
+
+	query[0] = type;
+	query[1] = upsname;
+	query[2] = varname;
+	numq = 3;
+
+	ret = upscli_get(&ups, numq, query, &numa, &answer);
+
+	if ((ret < 0) || (numa < numq))
+		return NULL;
+
+	/* <type> <upsname> <varname> <desc> */
+	return answer[3];
+}
+
+static void new_do_string(const char *varname, int maxlen)
+{
+	const	char	*val;
+
+	val = get_data("VAR", varname);
+
+	if (!val) {
+		printf("Unavailable\n");
+		fprintf(stderr, "new_do_string: can't get current value of %s\n",
+			varname);
+		return;
+	}
+
+	printf("<INPUT TYPE=\"TEXT\" NAME=\"UPSVAR_%s\" VALUE=\"%s\" "
+		"SIZE=\"%d\">\n", varname, val, maxlen);
+}
+
+static void new_do_enum(const char *varname)
+{
+	int	ret, numq, numa;
+	char	**answer, *val;
+	const	char	*query[4], *tmp;
+
+	/* get current value */
+	tmp = get_data("VAR", varname);
+
+	if (!tmp) {
+		printf("Unavailable\n");
+		fprintf(stderr, "do_enum: can't get current value of %s\n",
+			varname);
+		return;
+	}
+
+	/* tmp is a pointer into answer - have to save it somewhere else */
+	val = xstrdup(tmp);
+
+	query[0] = "ENUM";
+	query[1] = upsname;
+	query[2] = varname;
+	numq = 3;
+
+	ret = upscli_list_start(&ups, numq, query);
+
+	if (ret < 0) {
+		printf("Unavailable\n");
+		fprintf(stderr, "Error doing ENUM %s %s: %s\n", 
+			upsname, varname, upscli_strerror(&ups));
+		return;
+	}
+
+	ret = upscli_list_next(&ups, numq, query, &numa, &answer);
+
+	printf("<SELECT NAME=\"UPSVAR_%s\">\n", varname);
+
+	while (ret == 1) {
+
+		/* ENUM <upsname> <varname> <value> */
+
+		if (numa < 4) {
+			fprintf(stderr, "Error: insufficient data "
+				"(got %d args, need at least 4)\n", numa);
+
+			free(val);
+			return;
+		}
+
+		printf("<OPTION VALUE=\"%s\" ", answer[3]);
+
+		if (!strcmp(answer[3], val))
+			printf(" SELECTED");
+
+		printf(">%s</OPTION>\n", answer[3]);
+
+		ret = upscli_list_next(&ups, numq, query, &numa, &answer);
+	}
+
+	free(val);
+	printf("</SELECT>\n");
+}
+
+static void do_type(const char *varname)
+{
+	int	i, ret, numq, numa;
+	char	**answer;
+	const	char	*query[4];
+
+	query[0] = "TYPE";
+	query[1] = upsname;
+	query[2] = varname;
+	numq = 3;
+
+	ret = upscli_get(&ups, numq, query, &numa, &answer);
+
+	if ((ret < 0) || (numa < numq)) {
+		printf("Unknown type\n");	
+		return;
+	}
+
+	/* TYPE <upsname> <varname> <type>... */
+	for (i = 3; i < numa; i++) {
+
+		if (!strcasecmp(answer[i], "ENUM")) {
+			new_do_enum(varname);
+			return;
+		}
+
+		if (!strncasecmp(answer[i], "STRING:", 7)) {
+			char	*ptr, len;
+
+			/* split out the :<len> data */
+			ptr = strchr(answer[i], ':');
+			*ptr++ = '\0';
+			len = strtol(ptr, (char **) NULL, 10);
+
+			new_do_string(varname, len);
+			return;
+		}
+
+		/* ignore this one */
+		if (!strcasecmp(answer[i], "RW"))
+			continue;
+
+		printf("Unrecognized\n");
+	}
+}
+
+static void print_rw(const char *upsname, const char *varname)
+{
+	const	char	*tmp;
+
+	printf("<TR BGCOLOR=\"#60B0B0\" ALIGN=\"CENTER\">\n");
+
+	printf("<TD>");
+
+	tmp = get_data("DESC", varname);
+
+	if ((tmp) && (strcmp(tmp, "Unavailable") != 0))
+		printf("%s", tmp);
+	else
+		printf("%s", varname);
+
+	printf("</TD>\n");
+
+	printf("<TD>\n");
+	do_type(varname);
+	printf("</TD>\n");
+
+	printf("</TR>\n");
+}
+
+static void showsettings(void)
+{
+	int	ret, numq, numa;
+	const	char	*query[2];
+	char	**answer, *desc = NULL;
+	struct	list_t	*lhead, *llast, *ltmp, *lnext;
+
+	if (!checkhost(monups, &desc))
+		error_page("showsettings", "Access denied",
+		          "Access to that host is not authorized");
+
+	upsd_connect();
+
+	/* old default upsname scheme = use old protocol commands */
+	if ((!upsname) || (!strcmp(upsname, ""))) {
+		old_showsettings(desc);
+		return;
+	}
+
+	query[0] = "RW";
+	query[1] = upsname;
+	numq = 2;
+
+	ret = upscli_list_start(&ups, numq, query);
+
+	if (ret < 0) {
+
+		/* old upsd --> fall back on old LISTRW technique */
+		if (upscli_upserror(&ups) == UPSCLI_ERR_UNKCOMMAND) {
+			old_showsettings(desc);
+			return;
+		}
+
+		fprintf(stderr, "LIST RW %s failed: %s\n",
+			upsname, upscli_strerror(&ups));
+
+		error_page("showsettings", "Server protocol error",
+			"LIST RW command failed");
+
+		/* NOTREACHED */
+	}
+
+	llast = lhead = NULL;
+
+	ret = upscli_list_next(&ups, numq, query, &numa, &answer);
+
+	while (ret == 1) {
+
+		/* sock this entry away for later */
+
+		ltmp = xmalloc(sizeof(struct list_t));
+		ltmp->name = xstrdup(answer[2]);
+		ltmp->next = NULL;
+
+		if (llast)
+			llast->next = ltmp;
+		else
+			lhead = ltmp;
+
+		llast = ltmp;
+
+		ret = upscli_list_next(&ups, numq, query, &numa, &answer);
+	}
+
+	do_header("Current settings");
+	printf("<FORM ACTION=\"upsset.cgi\" METHOD=\"POST\">\n");
+	start_table();
+
+	/* include the description from checkhost() if present */
+        if (desc)
+		printf("<TR><TH BGCOLOR=\"#60B0B0\"COLSPAN=2>%s</TH></TR>\n",
+			desc);
+        
+	printf("<TR BGCOLOR=\"#60B0B0\">\n");
+	printf("<TH>Setting</TH>\n");
+	printf("<TH>Value</TH></TR>\n");
+
+	/* use the list to get descriptions and types */
+
+	ltmp = lhead;
+
+	while (ltmp) {
+		lnext = ltmp->next;
+
+		print_rw(upsname, ltmp->name);
+
+		free(ltmp->name);
+		free(ltmp);
+		ltmp = lnext;
+	}
+
+	printf("<TR BGCOLOR=\"#60B0B0\">\n");
+	printf("<TD COLSPAN=\"2\" ALIGN=\"CENTER\">\n");
+	do_hidden("savesettings");
+	printf("<INPUT TYPE=\"HIDDEN\" NAME=\"monups\" VALUE=\"%s\">\n", monups);
+	printf("<INPUT TYPE=\"SUBMIT\" VALUE=\"Save changes\">\n");
+	printf("<INPUT TYPE=\"RESET\" VALUE=\"Reset\">\n");
+	printf("</TD></TR>\n");
+	printf("</TABLE>\n");
+	printf("</FORM>\n");
+
+	printf("<TR><TD ALIGN=\"CENTER\">\n");
+	do_pickups("showsettings");
+	printf("</TD></TR>\n");
+
+	printf("</TABLE>\n");
+	printf("</BODY></HTML>\n");
+
+	upscli_disconnect(&ups);
+	exit(0);
+}
+
+static void old_setvar(const char *var, const char *val)
+{
+	char	buf[SMALLBUF];
+
+	upscli_getvar(&ups, upsname, var, buf, sizeof(buf));
+
+	/* don't send a SET if it hasn't chnaged */
+	if (strcmp(val, buf))
+		return;
+
+	printf("set %s to %s (was %s)\n", var, val, buf);
+
+	if (upsname)
+		snprintf(buf, sizeof(buf), "SET %s@%s %s\n", var, upsname, val);
+	else
+		snprintf(buf, sizeof(buf), "SET %s %s\n", var, val);
+
+	if (upscli_sendline(&ups, buf, strlen(buf)) < 0) {
+		printf("Error: SET failed: %s\n", upscli_strerror(&ups));
+		return;
+	}
+
+	if (upscli_readline(&ups, buf, sizeof(buf)) < 0) {
+		printf("Error: SET failed: %s\n", upscli_strerror(&ups));
+		return;
+	}
+
+	printf("response: %s\n", buf);
+}
+
+static void new_setvar(const char *var, const char *val)
+{
+	char	buf[SMALLBUF], enc[SMALLBUF];
+	const	char	*tmp;
+
+	/* get old value */
+	tmp = get_data("VAR", var);
+
+	if (!tmp) {
+		printf("Can't get old value for %s, aborting SET\n", var);
+		return;
+	}
+
+	/* don't send a SET if it hasn't chnaged */
+	if (!strcmp(tmp, var))
+		return;
+
+	printf("set %s to %s (was %s)\n", var, val, tmp);
+
+	snprintf(buf, sizeof(buf), "SET VAR %s %s \"%s\"\n",
+		upsname, var, pconf_encode(val, enc, sizeof(enc)));
+
+	if (upscli_sendline(&ups, buf, strlen(buf)) < 0) {
+		printf("Error: SET failed: %s\n", upscli_strerror(&ups));
+		return;
+	}
+
+	if (upscli_readline(&ups, buf, sizeof(buf)) < 0) {
+		printf("Error: SET failed: %s\n", upscli_strerror(&ups));
+		return;
+	}
+
+	if (strncmp(buf, "OK", 2) != 0) {
+		printf("Unexpected response: %s\n", buf);
+		return;
+	}
+
+	printf("OK\n");
+}
+
 /* turn a form submission of settings into SET commands for upsd */
 static void savesettings(void)
 {
-	char	*desc, buf[SMALLBUF];
+	char	*desc;
 	uvtype	*upsvar;
 
 	if (!checkhost(monups, &desc)) 
-		errorpage("Access denied",
-		          "Access to that host is not authorized",
-		          "showsettings");
+		error_page("showsettings", "Access denied",
+		          "Access to that host is not authorized");
 
 	upsd_connect();
 
 	upsvar = firstuv;
 
-	snprintf(buf, sizeof(buf), "USERNAME %s\n", username);
-
-	if (upscli_sendline(&ups, buf, strlen(buf)) < 0) {
-		printf("Send username failed\n");
-
-		fprintf(stderr, "Can't set username: %s\n",
-			upscli_strerror(&ups));
-		return;
-	}
-
-	if (upscli_readline(&ups, buf, sizeof(buf)) < 0) {
-		/* test for old upsd that doesn't do USERNAME */
-
-		if (ups.upserror == UPSCLI_ERR_UNKCOMMAND) {
-			errorpage("Protocol mismatch",
-			          "upsd version too old - USERNAME not supported",
-				  "showsettings");
-		}
-
-		errorpage("Can't set username", upscli_strerror(&ups),
-			   "showcmds");
-
-		upscli_disconnect(&ups);
-		exit(0);
-	}
-
-	snprintf(buf, sizeof(buf), "PASSWORD %s\n", password);
-	
-	if (upscli_sendline(&ups, buf, strlen(buf)) < 0) 
-		errorpage("Password set failed", upscli_strerror(&ups),
-		          "showcmds");
-
-	if (upscli_readline(&ups, buf, sizeof(buf)) < 0)
-		errorpage("Password set failed", upscli_strerror(&ups),
-		          "showcmds");
+	send_auth("showsettings");
 
 	do_header("Saving settings");
 	start_table();
@@ -849,32 +1434,22 @@ static void savesettings(void)
 	printf("<TR><TD><PRE>\n");
 
 	while (upsvar) {
-		upscli_getvar(&ups, upsname, upsvar->var, buf, sizeof(buf));
 
-		if (strcmp(upsvar->value, buf)) {
-			printf("set %s to %s (was %s)\n", upsvar->var, 
-			        upsvar->value, buf);
-
-			if (upsname)
-				snprintf(buf, sizeof(buf), "SET %s@%s %s\n", 
-				         upsvar->var, upsname, upsvar->value);
-			else
-				snprintf(buf, sizeof(buf), "SET %s %s\n", 
-				         upsvar->var, upsvar->value);
-
-			if (upscli_sendline(&ups, buf, strlen(buf)) < 0) {
-				printf("Error: SET failed: %s\n",
-				       upscli_strerror(&ups));
-			} else {
-				if (upscli_readline(&ups, buf, sizeof(buf)) < 0) {
-					printf("Error: SET failed: %s\n",
-					       upscli_strerror(&ups));
-				} else {
-					printf("response: %s\n", buf);
-				}
-			}
+		/* fallback on old methods */
+		if (!upsname) {
+			old_setvar(upsvar->var, upsvar->value);
+			upsvar = upsvar->next;
+			continue;
 		}
 
+		/* ditto for old variable names */
+		if (!strchr(upsvar->var, '.')) {
+			old_setvar(upsvar->var, upsvar->value);
+			upsvar = upsvar->next;
+			continue;
+		}
+
+		new_setvar(upsvar->var, upsvar->value);
 		upsvar = upsvar->next;
 	}
 
@@ -888,6 +1463,7 @@ static void savesettings(void)
 	printf("</TD></TR></TABLE>\n");
 	printf("</BODY></HTML>\n");
 
+	upscli_disconnect(&ups);
 	exit(0);
 }
 
@@ -903,37 +1479,29 @@ static void initial_pickups(void)
 	printf("</TABLE>\n");
 	printf("</TD></TR></TABLE>\n");
 	printf("</BODY></HTML>\n");
+
+	upscli_disconnect(&ups);
 	exit(0);
 }
 
-void upsset_conf_arg(int numargs, char **arg)
+void upsset_conf_err(const char *errmsg)
 {
-	if (numargs < 1)
-		return;
-
-	if (!strcmp(arg[0], MAGIC_ENABLE_STRING))
-		magic_string_set = 1;
+	upslogx(LOG_ERR, "Fatal error in parseconf(upsset.conf): %s", errmsg);
 }
 
-void upsset_conf_err(int linenum, char *errtext)
-{
-	if (linenum == 0)
-		fprintf(stderr, "upsset.conf: %s\n", errtext);
-	else
-		fprintf(stderr, "upsset.conf line %d: %s\n", linenum, errtext);
-} 
-
 /* see if the user has confirmed their cgi directory's secure state */
 static void check_conf(void)
 {
 	char	fn[SMALLBUF];
-	int	ret;
+	PCONF_CTX	ctx;
 
-	snprintf(fn, sizeof(fn), "%s/upsset.conf", CONFPATH);
+	snprintf(fn, sizeof(fn), "%s/upsset.conf", confpath());
 
-	ret = parseconf(fn, upsset_conf_arg, upsset_conf_err);
+	pconf_init(&ctx, upsset_conf_err);
+
+	if (!pconf_file_begin(&ctx, fn)) {
+		pconf_finish(&ctx);
 
-	if (ret != 0) {
 		printf("<PRE>\n");
 		printf("Error: Can't open upsset.conf to verify security settings.\n");
 		printf("Refusing to start until this is fixed.\n");
@@ -944,6 +1512,22 @@ static void check_conf(void)
 		exit(0);
 	}
 
+	while (pconf_file_next(&ctx)) {
+		if (pconf_parse_error(&ctx)) {
+			upslogx(LOG_ERR, "Parse error: %s:%d: %s",
+				fn, ctx.linenum, ctx.errmsg);
+			continue;
+		}
+
+		if (ctx.numargs < 1)
+			continue;
+
+		if (!strcmp(ctx.arglist[0], MAGIC_ENABLE_STRING))
+			magic_string_set = 1;
+	}
+
+	pconf_finish(&ctx);
+
 	/* if we've been enabled, jump out of here and go to work */
 	if (magic_string_set == 1)
 		return;
