@@ -65,7 +65,7 @@ void parsearg(char *var, char *value)
 
 static void report_error(void)
 {
-	if (ups.upserror == UPSCLI_ERR_VARNOTSUPP)
+	if (upscli_upserror(&ups) == UPSCLI_ERR_VARNOTSUPP)
 		printf("Not supported\n");
 	else
 		printf("[error: %s]\n", upscli_strerror(&ups));
@@ -74,7 +74,7 @@ static void report_error(void)
 /* make sure we're actually connected to upsd */
 static int check_ups_fd(int do_report)
 {
-	if (ups.fd == -1) {
+	if (upscli_fd(&ups) == -1) {
 		if (do_report)
 			report_error();
 
@@ -164,10 +164,14 @@ static int get_img_val(const char *var, const char *desc, const char *imgargs)
 		return 1;
 	}
 
-	printf("<IMG SRC=\"upsimage.cgi?host=%s&amp;display=%s&amp;%s\""
-		" WIDTH=\"100\" HEIGHT=\"350\""
-		" ALT=\"%s: %s\">\n",
-			currups->sys, var, imgargs, desc, temp);
+	printf("<IMG SRC=\"upsimage.cgi?host=%s&amp;display=%s",
+		currups->sys, var);
+
+	if ((imgargs) && (strlen(imgargs) > 0))
+		printf("&amp;%s", imgargs);
+
+	printf("\" WIDTH=\"100\" HEIGHT=\"350\" ALT=\"%s: %s\">\n",
+		desc, temp);
 
 	return 1;
 }
@@ -238,6 +242,9 @@ static int do_img(char *buf)
 		split_imgarg(ptr, imgargs, sizeof(imgargs));
 	}
 
+	if (!strcmp(type, "BATTVOLT"))
+		return get_img_val("battvolt", "Battery Voltage", imgargs);
+
 	if (!strcmp(type, "BATTPCT"))
 		return get_img_val("battpct", "Battery Charge", imgargs);
 
@@ -262,13 +269,33 @@ static int do_img(char *buf)
 static void ups_connect(void)
 {
 	static ulist_t	*lastups = NULL;
+	char	*newups, *newhost;
+	int	newport;
 
-	/* don't reconnect if these are both on the same host */
+	/* try to minimize reconnects */
 	if (lastups) {
+
+		/* don't reconnect if these are both the same UPS */
 		if (!strcmp(lastups->sys, currups->sys)) {
 			lastups = currups;
 			return;
 		}
+
+		/* see if it's just on the same host */
+		upscli_splitname(currups->sys, &newups, &newhost, &newport);
+
+		if ((!strcmp(newhost, hostname)) && (port == newport)) {
+			free(upsname);
+			upsname = newups;
+
+			free(newhost);
+			lastups = currups;
+			return;
+		}
+
+		/* not the same upsd, so disconnect */
+		free(newups);
+		free(newhost);
 	}
 
 	upscli_disconnect(&ups);
@@ -594,7 +621,7 @@ static void display_template(const char *tfn)
 {
 	char	fn[SMALLBUF], buf[LARGEBUF];	
 
-	snprintf(fn, sizeof(fn), "%s/%s", CONFPATH, tfn);
+	snprintf(fn, sizeof(fn), "%s/%s", confpath(), tfn);
 
 	tf = fopen(fn, "r");
 
@@ -640,38 +667,54 @@ static void add_ups(char *sys, char *desc)
 		ulhead = tmp;
 }
 
-void hosts_conf_args(int numargs, char **arg)
+/* called for fatal errors in parseconf like malloc failures */
+void upsstats_hosts_err(const char *errmsg)
 {
-	if (numargs < 3)
-		return;
-
-	/* MONITOR <host> <desc> */
-	if (!strcmp(arg[0], "MONITOR"))
-		add_ups(arg[1], arg[2]);
-}
-
-void hosts_conf_error(int linenum, char *errtext)
-{
-	/* this file must exist ... */
-	if (linenum == 0) {
-		printf("</TABLE></TD></TR></TABLE>\n");
-		printf("Error: can't open hosts.conf\n");
-		printf("</BODY></HTML>\n");
-
-		/* leave something for the admin */
-		fprintf(stderr, "hosts.conf: %s\n", errtext);
-		exit(0);
-	}
-
-	fprintf(stderr, "hosts.conf line %d: %s\n", linenum, errtext);
+	upslogx(LOG_ERR, "Fatal error in parseconf(hosts.conf): %s", errmsg);
 }
 
 static void load_hosts_conf(void)
 {
 	char	fn[SMALLBUF];
+	PCONF_CTX	ctx;
 
 	snprintf(fn, sizeof(fn), "%s/hosts.conf", CONFPATH);
-	parseconf(fn, hosts_conf_args, hosts_conf_error);
+
+	pconf_init(&ctx, upsstats_hosts_err);
+
+	if (!pconf_file_begin(&ctx, fn)) {
+		pconf_finish(&ctx);
+
+		printf("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\"\n");
+		printf("	\"http://www.w3.org/TR/REC-html40/loose.dtd\">\n");
+		printf("<HTML><HEAD>\n");
+		printf("<TITLE>Error: can't open hosts.conf</TITLE>\n");
+		printf("</HEAD><BODY>\n");
+		printf("Error: can't open hosts.conf\n");
+		printf("</BODY></HTML>\n");
+
+		/* leave something for the admin */
+		fprintf(stderr, "upsstats: %s\n", ctx.errmsg);
+		exit(0);
+	}
+
+	while (pconf_file_next(&ctx)) {
+		if (pconf_parse_error(&ctx)) {
+			upslogx(LOG_ERR, "Parse error: %s:%d: %s",
+				fn, ctx.linenum, ctx.errmsg);
+			continue;
+		}
+
+		if (ctx.numargs < 3)
+			continue;
+
+		/* MONITOR <host> <desc> */
+		if (!strcmp(ctx.arglist[0], "MONITOR"))
+			add_ups(ctx.arglist[1], ctx.arglist[2]);
+
+	}
+
+	pconf_finish(&ctx);
 }
 
 static void display_single(void)
