@@ -49,6 +49,21 @@ void syslogbit_set(void)
 	xbit_set(&upslog_flags, UPSLOG_SYSLOG);
 }
 
+/* get the syslog ready for us */
+void open_syslog(const char *progname)
+{
+	int	opt;
+
+	opt = LOG_PID;
+
+	/* we need this to grab /dev/log before chroot */
+#ifdef LOG_NDELAY
+	opt |= LOG_NDELAY;
+#endif
+
+	openlog(progname, opt, LOG_FACILITY);
+}
+
 /* close ttys and become a daemon */
 void background(void)
 {
@@ -82,43 +97,46 @@ void background(void)
 	upslogx(LOG_INFO, "Startup successful");
 }
 
-
-void switch_to_user(const char *user, const char *group)
+/* do this here to keep pwd/grp stuff out of the main files */
+struct passwd *get_user_pwent(const char *name)
 {
-	struct passwd *pw;
-	struct group  *gr;
-	gid_t gid;
+	return getpwnam(name);
+}
+
+/* change to the user defined in the struct */
+void become_user(struct passwd *pw)
+{
+	/* if we can't switch users, then don't even try */
+	if ((geteuid() != 0) && (getuid() != 0))
+		return;
 
 	if (getuid() == 0)
 		if (seteuid(0))
-			; /* error */
+			fatal("getuid gave 0, but seteuid(0) failed");
 
-	if ((pw = getpwnam(user)) == NULL)
-		fatalx("getpwnam() failed: user %s does not exist", user);
-	gid = pw->pw_gid;
-
-	if (initgroups(user, pw->pw_gid) == -1)
+	if (initgroups(pw->pw_name, pw->pw_gid) == -1)
 		fatal("initgroups");
 
-	if (group) {
-		if ((gr = getgrnam(group)) == NULL)
-			fatalx("getgrnam() failed: group %s does not exist", group);
-		gid = gr->gr_gid;
-	}
-
-	if (setgid(gid) == -1)
+	if (setgid(pw->pw_gid) == -1)
 		fatal("setgid");
 
-        /* change uid second */
 	if (setuid(pw->pw_uid) == -1)
 		fatal("setuid");
 }
 
-/* change uid/gid if running as root */
-void droproot(void)
+/* drop down into a directory and throw away pointers to the old path */
+void chroot_start(const char *path)
 {
-	if (getuid() == 0 || geteuid() == 0)
-		switch_to_user(RUN_AS_USER, RUN_AS_GROUP);
+	if (chdir(path))
+		fatal("chdir(%s)", path);
+
+	if (chroot(path))
+		fatal("chroot(%s)", path);
+
+	if (chdir("/"))
+		fatal("chdir(/)");
+
+	upsdebugx(1, "chrooted into %s", path);
 }
 
 /* drop off a pidfile for this process */
@@ -223,18 +241,6 @@ const char *xbasename(const char *file)
 	return p + 1;
 }
 
-int findinfo(const itype *info, int size, int offset, int type)
-{
-	int x;
-
-	for (x = offset; x < size; x++) {
-		if (info[x].type == type)
-			return x;
-	}
-
-	return -1;
-}
-
 void vupslog(int priority, const char *fmt, va_list va, int use_strerror)
 {
 	char buf[LARGEBUF];
@@ -251,6 +257,38 @@ void vupslog(int priority, const char *fmt, va_list va, int use_strerror)
 		syslog(priority, "%s", buf);
 }
 
+/* Return the default path for the directory containing configuration files */
+const char * confpath(void) 
+{
+	const char * path;
+
+	if ((path = getenv("NUT_CONFPATH")) == NULL)
+		path = CONFPATH;
+
+	return path;
+}
+
+/* Return the default path for the directory containing state files */
+const char * dflt_statepath(void) 
+{
+	const char * path;
+
+	if ((path = getenv("NUT_STATEPATH")) == NULL)
+		path = STATEPATH;
+
+	return path;
+}
+
+/* Return the alternate path for pid files */
+const char * altpidpath(void) 
+{
+#ifdef ALTPIDPATH
+	return ALTPIDPATH;
+#else
+	return dflt_statepath();
+#endif
+}
+
 /* logs the formatted string to any configured logging devices + the output of strerror(errno) */
 void upslog(int priority, const char *fmt, ...)
 {
