@@ -29,7 +29,7 @@
 	static	char	*ups_section;
 
 /* handle arguments separated by parseconf */
-void conf_args(int numargs, char **arg)
+static void conf_args(int numargs, char **arg)
 {
 	char	*ep;
 
@@ -72,28 +72,41 @@ void conf_args(int numargs, char **arg)
 	}
 }
 
-void conf_err(int linenum, char *errtxt)
+/* called for fatal errors in parseconf like malloc failures */
+void upsconf_err(const char *errmsg)
 {
-	if (linenum == 0)
-		upslogx(LOG_ERR, "ups.conf: %s", errtxt);
-	else
-		upslogx(LOG_ERR, "ups.conf line %d: %s", errtxt);
+        upslogx(LOG_ERR, "Fatal error in parseconf(ups.conf): %s", errmsg);
 }
 
 /* open the ups.conf, parse it, and call back do_upsconf_args() */
 void read_upsconf(int required)
 {
-	int	ret;
 	char	fn[SMALLBUF];
-
-	snprintf(fn, sizeof(fn), "%s/ups.conf", CONFPATH);
+	PCONF_CTX	ctx;
 
 	ups_section = NULL;
+	snprintf(fn, sizeof(fn), "%s/ups.conf", confpath());
 
-	ret = parseconf(fn, conf_args, conf_err);
+	pconf_init(&ctx, upsconf_err);
 
-	if ((ret != 0) && (required))
-		fatalx("ups.conf parsing failed");
+	if (!pconf_file_begin(&ctx, fn)) {
+		if (required)
+			fatalx("ups.conf parsing failed: %s", ctx.errmsg);
+
+		return;
+	}
+
+	while (pconf_file_next(&ctx)) {
+		if (pconf_parse_error(&ctx)) {
+			upslogx(LOG_ERR, "Parse error: %s:%d: %s",
+				fn, ctx.linenum, ctx.errmsg);
+			continue;
+		}
+
+		conf_args(ctx.numargs, ctx.arglist);
+	}
+
+	pconf_finish(&ctx);
 
 	if (ups_section)
 		free(ups_section);
