@@ -20,6 +20,8 @@
 #include "main.h"
 #include "apcsmart.h"
 
+#define DRV_VERSION "0.72"
+
 	extern	int	sddelay;
 	int	cap_measureupsii = 0, upslevel = 0;
 	int	shutdowncmdset = 0; /* The shutdown command set to use */
@@ -27,73 +29,73 @@
 #define ALT_CABLE_1 "940-0095B"
 
 /* try to find AMBTEMP, CONTACTS and AMBHUMID */
-void test_measureupsii(void)
+static void test_measureupsii(void)
 {
+	int	ret;
 	char	temp[256];
 
 	upssendchar(REQ_AMBTEMP);
-	upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
 
-	if (!strcmp(temp, "NA"))	/* not supported */
+	if ((ret < 1) || (!strcmp(temp, "NA")))	/* not supported */
 		return;
 
 	/* we got something, so it must be supported */
 
 	cap_measureupsii = 1;		/* log this fact */
 
-	addinfo(INFO_AMBTEMP, temp, 0, 0);
+	dstate_setinfo("ambient.temperature", "%s", temp);
 
 	/* Now probe the contacts */
 
 	upssendchar(REQ_CONTACTS);
-	upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
 
-	addinfo(INFO_CONTACTS, temp, 0, 0);
+	if ((ret > 0) && (strcmp(temp, "NA") != 0))
+		dstate_setinfo("ups.contacts", "%s", temp);
 
 	/* now try for ambient humidity support */
 
 	upssendchar(REQ_AMBHUMID);
-	upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
 
-	if (!strcmp(temp, "NA"))	/* not supported */
-		return;
-
-	addinfo(INFO_AMBHUMID, temp, 0, 0);
+	if ((ret > 0) && (strcmp(temp, "NA") != 0))
+		dstate_setinfo("ambient.humidity", "%s", temp);
 }
 
-void addcap(char cmdchar, char *value)
+static void addcap(char cmdchar, const char *value)
 {
 	switch (cmdchar) {
-		case 'u': addenum(INFO_HIGHXFER, value);
-			  setflag(INFO_HIGHXFER, FLAG_RW | FLAG_ENUM);
+		case 'u': dstate_addenum("input.transfer.high", "%s", value);
+			  dstate_setflags("input.transfer.high", ST_FLAG_RW);
 			  break;
 
-		case 'l': addenum(INFO_LOWXFER, value);
-			  setflag(INFO_LOWXFER, FLAG_RW | FLAG_ENUM);
+		case 'l': dstate_addenum("input.transfer.low", "%s", value);
+			  dstate_setflags("input.transfer.low", ST_FLAG_RW);
 			  break;
 
-		case 'e': addenum(INFO_RTHRESH, value);
-			  setflag(INFO_RTHRESH, FLAG_RW | FLAG_ENUM);
+		case 'e': dstate_addenum("battery.charge.restart", "%s", value);
+			  dstate_setflags("battery.charge.restart", ST_FLAG_RW);
 			  break;
 
 		case 'o': break;	/* output voltages, unused now	*/
 
-		case 's': addenum(INFO_LINESENS, value);
-			  setflag(INFO_LINESENS, FLAG_RW | FLAG_ENUM);
+		case 's': dstate_addenum("input.sensitivity", "%s", value);
+			  dstate_setflags("input.sensitivity", ST_FLAG_RW);
 			  break;
 
 		case 'q': break;	/* TODO: low batt warning	*/
 
-		case 'p': addenum(INFO_GRACEDELAY, value);
-			  setflag(INFO_GRACEDELAY, FLAG_RW | FLAG_ENUM);
+		case 'p': dstate_addenum("ups.delay.shutdown", "%s", value);
+			  dstate_setflags("ups.delay.shutdown", ST_FLAG_RW);
 			  break;
 
-		case 'k': addenum(INFO_ALRMDEL, value);
-			  setflag(INFO_ALRMDEL, FLAG_RW | FLAG_ENUM);
+		case 'k': dstate_addenum("battery.alarm.threshold", "%s", value);
+			  dstate_setflags("battery.alarm.threshold", ST_FLAG_RW);
 			  break;
 
-		case 'r': addenum(INFO_WAKEDELAY, value);
-			  setflag(INFO_WAKEDELAY, FLAG_RW | FLAG_ENUM);
+		case 'r': dstate_addenum("ups.delay.start", "%s", value);
+			  dstate_setflags("ups.delay.start", ST_FLAG_RW);
 			  break;
 
 		case 'E': break;	/* TODO: selftest intervals	*/
@@ -106,17 +108,38 @@ void addcap(char cmdchar, char *value)
 	}		
 }	
 
-void do_capabilities(void)
+static void poll_data(const char *var, char reqchar)
+{
+	int	ret;
+	char	tmp[SMALLBUF];
+
+	upssendchar(reqchar);
+	ret = upsrecv(tmp, sizeof(tmp), ENDCHAR, IGNCHARS);
+
+	if ((ret < 0) || (!strcmp(tmp, "NA")))
+		dstate_delinfo(var);
+	else
+		dstate_setinfo(var, "%s", tmp);
+}
+
+static void do_capabilities(void)
 {
 	char	upsloc, temp[512], *ptr, cmd, loc, *entptr, etmp[16], *endtemp;
-	int	nument, entlen, i, matrix, recvret;
+	int	nument, entlen, i, matrix, ret;
 
 	/* just in case ... */
 	usleep(50000);
 
 	/* get location byte for later comparisons */
 	upssendchar('V');
-	upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+	if ((ret < 1) || (!strcmp(temp, "NA"))) {
+		upslogx(LOG_ERR, "Get location byte failed");
+		return;
+	}
+
 	upsloc = temp[strlen(temp) - 1];
 
 	/* add base entries for these variables */
@@ -124,31 +147,28 @@ void do_capabilities(void)
 	/* create them as normal variables for now - enum and rw flags    *
 	 * get added later if the model actually supports such operations */
 
-	addinfo(INFO_LOWXFER, "", 0, 0);
-	addinfo(INFO_HIGHXFER, "", 0, 0);
-	addinfo(INFO_WAKEDELAY, "", 0, 0);
-	addinfo(INFO_LINESENS, "", 0, 0);
-	addinfo(INFO_GRACEDELAY, "", 0, 0);
-	addinfo(INFO_RTHRESH, "", 0, 0);
-	addinfo(INFO_ALRMDEL, "", 0, 0);
-	installinfo(INFO_LOWXFER, REQ_LOWXFER, ENDCHAR, IGNCHARS);
-	installinfo(INFO_HIGHXFER, REQ_HIGHXFER, ENDCHAR, IGNCHARS);
-	installinfo(INFO_WAKEDELAY, REQ_WAKEDELAY, ENDCHAR, IGNCHARS);
-	installinfo(INFO_LINESENS, REQ_LINESENS, ENDCHAR, IGNCHARS);
-	installinfo(INFO_GRACEDELAY, REQ_GRACEDELAY, ENDCHAR, IGNCHARS);
-	installinfo(INFO_RTHRESH, REQ_RTHRESH, ENDCHAR, IGNCHARS);
-	installinfo(INFO_ALRMDEL, REQ_ALRMDEL, ENDCHAR, IGNCHARS);
+	poll_data("input.transfer.low", REQ_LOWXFER);
+	poll_data("input.transfer.high", REQ_HIGHXFER);
+	poll_data("ups.delay.start", REQ_WAKEDELAY);
+	poll_data("input.sensitivity", REQ_LINESENS);
+	poll_data("ups.delay.shutdown", REQ_GRACEDELAY);
+	poll_data("battery.charge.restart", REQ_RTHRESH);
+	poll_data("battery.alarm.threshold", REQ_ALRMDEL);
 
 	/* battery change date */
-	addinfo(INFO_BATTDATE, "", FLAG_RW | FLAG_STRING, 8);
-	installinfo(INFO_BATTDATE, REQ_BATTDATE, ENDCHAR, IGNCHARS);
+	poll_data("battery.date", REQ_BATTDATE);
+
+	/* writable string, max 8 bytes */
+	dstate_setflags("battery.date", ST_FLAG_RW | ST_FLAG_STRING);
+	dstate_setaux("battery.date", 8);
 
 	/* get capability string */
 	upssendchar(REQ_CAPABILITIES);	/* ^Z */
-	strcpy(temp, "NA");
-	recvret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
 
-	if ((recvret == -1) || (!strcmp(temp, "NA"))) {
+	snprintf(temp, sizeof(temp), "NA");
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+	if ((ret == -1) || (!strcmp(temp, "NA"))) {
 		/* Early Smart-UPS, not as smart as later ones */
 		return;
 	}
@@ -201,18 +221,23 @@ void do_capabilities(void)
 	}
 }
 
-void do_status(void)
+static int do_status(void)
 {
 	char	temp[256];
-	int	tval;
+	int	ret, tval;
 
 	/* parse the status */
 	upssendchar(REQ_STATUS);
-	upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+	if ((ret < 1) || (!strcmp(temp, "NA"))) {
+		dstate_datastale();
+		return 0;
+	}
 
 	/* APC SU600 models return this when the FPTEST is still active */
 	if (!strncmp(temp, "BC", 2))
-		return;			/* bogus, we don't want to parse it */
+		return 1;		/* bogus, we don't want to parse it */
 
 	tval = strtol(temp, 0, 16);
 
@@ -239,57 +264,60 @@ void do_status(void)
 		status_set("OFF");
 
 	status_commit();
+	dstate_dataok();
+
+	return 1;
 }
 
-void getbaseinfo(const char *port)
+static void getbaseinfo(const char *port)
 {
 	char	temp[256];
 	int	silentfail = 0, oldmatrix = 0;
-	int	recvret = 0;
+	int	ret = 0;
 
 	upslevel = 1;
 
 	/* really old models ignore REQ_MODEL, so find them first */
 	upssendchar(REQ_MODEL);
-	strcpy(temp, "NA");
+	snprintf(temp, sizeof(temp), "NA");
 
 	/* disable timeout complaints temporarily */
 	flag_timeoutfailure = -1;
-	recvret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
 	flag_timeoutfailure = 0;
 
-	if ((recvret == -1) || (!strcmp(temp, "NA"))) {
+	if ((ret == -1) || (!strcmp(temp, "NA"))) {
 		/* force the model name */
-		addinfo(INFO_MODEL, "Smart-UPS", 0, 0);
+		dstate_setinfo("ups.model", "Smart-UPS");
 		silentfail = 1;
 	}
 	else
-		addinfo(INFO_MODEL, temp, 0, 0);
+		dstate_setinfo("ups.model", "%s", temp);
 
 	/* see if this might be an old Matrix-UPS instead */
 	upssendchar('/');
 	flag_timeoutfailure = -1;
-	strcpy(temp, "NA");
-	recvret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+	snprintf(temp, sizeof(temp), "NA");
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
 	flag_timeoutfailure = 0;
 
 	/* it returned a useful number, so it must be a Matrix-UPS */
-	if ((recvret != -1) && (strcmp(temp, "NA")) != 0) {
+	if ((ret != -1) && (strcmp(temp, "NA")) != 0) {
 		oldmatrix = 1;	/* doesn't do 'b' */
-		setinfo(INFO_MODEL, "%s", "Matrix-UPS");
+		dstate_setinfo("ups.model", "Matrix-UPS");
 	}
 
-	addinfo(INFO_SERIAL, "", 0, 0);
-	installinfo(INFO_SERIAL, REQ_SERIAL, ENDCHAR, IGNCHARS);
+	poll_data("ups.serial", REQ_SERIAL);
 
-	addinfo(INFO_STATUS, "", 0, 0);
 	do_status();
 
 	/* upslevel 1: Smart-UPS v/s, older Back-UPS Pros */
 	upssendchar(REQ_UTILITY);
-	strcpy(temp, "NA");
-	recvret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
-	if ((!strcmp(temp, "NA")) || (recvret == -1)) {
+
+	snprintf(temp, sizeof(temp), "NA");
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+	if ((!strcmp(temp, "NA")) || (ret == -1)) {
 		upslevel = 1;
 		return;
 	}
@@ -299,46 +327,39 @@ void getbaseinfo(const char *port)
 
 	/* add other things to monitor and prime them with values */
 
-	addinfo(INFO_UTILITY, "", 0, 0);
-	installinfo(INFO_UTILITY, REQ_UTILITY, ENDCHAR, IGNCHARS);
-
-	addinfo(INFO_BATTPCT, "", 0, 0);
-	installinfo(INFO_BATTPCT, REQ_BATTPCT, ENDCHAR, IGNCHARS);
-
-	addinfo(INFO_ACFREQ, "", 0, 0);
-	installinfo(INFO_ACFREQ, REQ_ACFREQ, ENDCHAR, IGNCHARS);
-
-	addinfo(INFO_LOADPCT, "", 0, 0);
-	installinfo(INFO_LOADPCT, REQ_LOADPCT, ENDCHAR, IGNCHARS);
-
-	addinfo(INFO_BATTVOLT, "", 0, 0);
-	installinfo(INFO_BATTVOLT, REQ_BATTVOLT, ENDCHAR, IGNCHARS);
-
-	addinfo(INFO_OUTVOLT, "", 0, 0);
-	installinfo(INFO_OUTVOLT, REQ_OUTVOLT, ENDCHAR, IGNCHARS);
+	poll_data("input.voltage", REQ_UTILITY);
+	poll_data("input.frequency", REQ_ACFREQ);
+	poll_data("ups.load", REQ_LOADPCT);
+	poll_data("battery.charge", REQ_BATTPCT);
+	poll_data("battery.voltage", REQ_BATTVOLT);
+	poll_data("output.voltage", REQ_OUTVOLT);
 
 	/* upslevel 3: at least a Smart-UPS (beyond the Back-UPS Pro) */
 
 	/* test for UPSTEMP support */
-	strcpy(temp, "NA");
+	snprintf(temp, sizeof(temp), "NA");
 
 	upssendchar(REQ_UPSTEMP);
-	upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
-	if (strcmp(temp, "NA") != 0) {
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+	if ((ret > 0) && (strcmp(temp, "NA") != 0)) {
 		upslevel = 3;
-		addinfo(INFO_UPSTEMP, temp, 0, 0);
+		dstate_setinfo("ups.temperature", "%s", temp);
 	}
 
-	addinfo(INFO_UPSIDENT, "", FLAG_RW | FLAG_STRING, 8);
-	installinfo(INFO_UPSIDENT, REQ_UPSIDENT, ENDCHAR, IGNCHARS);
+	poll_data("ups.id", REQ_UPSIDENT);
+
+	/* another 8 char rw string */
+	dstate_setflags("ups.id", ST_FLAG_RW | ST_FLAG_STRING);
+	dstate_setaux("ups.id", 8);
 
 	/* now add the instant commands */
-	addinfo(INFO_INSTCMD, "", 0, CMD_FPTEST);
-	addinfo(INFO_INSTCMD, "", 0, CMD_BTEST1);
-	addinfo(INFO_INSTCMD, "", 0, CMD_OFF);
-	addinfo(INFO_INSTCMD, "", 0, CMD_ON);
-	addinfo(INFO_INSTCMD, "", 0, CMD_CAL0);
-	addinfo(INFO_INSTCMD, "", 0, CMD_CAL1);
+	dstate_addcmd("test.panel.start");
+	dstate_addcmd("test.battery.start");
+	dstate_addcmd("load.off");
+	dstate_addcmd("load.on");
+	dstate_addcmd("calibrate.start");
+	dstate_addcmd("calibrate.stop");
 
 	do_capabilities();
 
@@ -350,51 +371,66 @@ void getbaseinfo(const char *port)
 	test_measureupsii();
 }
 
+/* get the UPS talking to us in smart mode */
+static int smartmode(void)
+{
+	int	tries, ret;
+	char	temp[256];
+
+	for (tries = 0; tries < 5; tries++) {
+		upssendchar(GO_SMART);
+		ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+		if ((ret > 0) && (!strcmp(temp, "SM")))
+			return 1;	/* success */
+
+		sleep(1);	/* wait before trying again */
+	}
+
+	return 0;	/* failure */
+}
+
 /* normal idle loop - keep up with the current state of the UPS */
 void upsdrv_updateinfo(void)
 {
-	char	temp[256];
-
 	/* try to wake up a dead ups once in awhile */
 	if (flag_timeoutfailure == 1) {
-		upssendchar(GO_SMART);
-		upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+		if (!smartmode) {
+			dstate_datastale();
+			return;
+		}
 	}
 
-	do_status();
+	if (!do_status())
+		return;
 
 	/* upslevel 1 only has a status readback */
-	if (upslevel < 2) {
-		writeinfo();
+	if (upslevel < 2)
 		return;
-	}
 
-	installinfo(INFO_UTILITY, REQ_UTILITY, ENDCHAR, IGNCHARS);
-	installinfo(INFO_BATTPCT, REQ_BATTPCT, ENDCHAR, IGNCHARS);
-	installinfo(INFO_ACFREQ, REQ_ACFREQ, ENDCHAR, IGNCHARS);
-	installinfo(INFO_LOADPCT, REQ_LOADPCT, ENDCHAR, IGNCHARS);
-	installinfo(INFO_BATTVOLT, REQ_BATTVOLT, ENDCHAR, IGNCHARS);
-	installinfo(INFO_OUTVOLT, REQ_OUTVOLT, ENDCHAR, IGNCHARS);
+	poll_data("input.voltage", REQ_UTILITY);
+	poll_data("input.frequency", REQ_ACFREQ);
+	poll_data("ups.load", REQ_LOADPCT);
+	poll_data("battery.charge", REQ_BATTPCT);
+	poll_data("battery.voltage", REQ_BATTVOLT);
+	poll_data("output.voltage", REQ_OUTVOLT);
 
 	/* need at least 3 for ups temperature and Measure-UPS II */
-	if (upslevel < 3) {
-		writeinfo();
+	if (upslevel < 3)
 		return;
-	}
 
-	installinfo(INFO_UPSTEMP, REQ_UPSTEMP, ENDCHAR, IGNCHARS);
+	poll_data("ups.temperature", REQ_UPSTEMP);
 
 	if (cap_measureupsii) {
-		installinfo(INFO_AMBHUMID, REQ_AMBHUMID, ENDCHAR, IGNCHARS);
-		installinfo(INFO_AMBTEMP, REQ_AMBTEMP, ENDCHAR, IGNCHARS);
-		installinfo(INFO_CONTACTS, REQ_CONTACTS, ENDCHAR, IGNCHARS);
+		poll_data("ambient.humidity", REQ_AMBHUMID);
+		poll_data("ambient.temperature", REQ_AMBTEMP);
+		poll_data("ups.contacts", REQ_CONTACTS);
 	}
-
-	writeinfo();
 }
 
 /* TODO: roll this into upscommon ala bestups */
-void sendstring(char * string, int len, int delay)
+static void sendstring(char * string, int len, int delay)
 {
 	int	i;
 
@@ -404,33 +440,8 @@ void sendstring(char * string, int len, int delay)
 	}
 }
 
-/* get the UPS talking to us in smart mode */
-int smartmode(void)
-{
-	int	tries;
-	char	temp[256];
-
-	tries = 0;
-	for (;;) {
-		tries++;
-
-		upssendchar(GO_SMART);
-		upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
-
-		if (!strcmp(temp, "SM"))
-			return 1;	/* success */
-
-		if (tries > 5)
-			return 0;	/* failure */
-
-		sleep(1);	/* wait before trying again */
-	}
-
-	return 0;	/* notreached */
-}
-
 /* 940-0095B support: set DTR, lower RTS */
-void init_serial_0095B(void)
+static void init_serial_0095B(void)
 {
 	int	dtr_bit = TIOCM_DTR;
 	int	rts_bit = TIOCM_RTS;
@@ -440,13 +451,16 @@ void init_serial_0095B(void)
 }
 
 /* set a value in the hardware using the <cmdchar> '-' (repeat) approach */
-void hw_set (char cmdchar, char *newval)
+static void hw_set(char cmdchar, const char *newval)
 {
 	char	temp[256], orig[256];
-	int	tries;
+	int	tries, ret;
 
 	upssendchar(cmdchar);
-	upsrecv(orig, sizeof(orig), ENDCHAR, IGNCHARS);
+	ret = upsrecv(orig, sizeof(orig), ENDCHAR, IGNCHARS);
+
+	if ((ret < 1) || (!strcmp(orig, "NA")))
+		return;
 
 	/* don't change it if already set - save wear and tear on the eeprom */
 	if (!strcmp(orig, newval))
@@ -455,11 +469,16 @@ void hw_set (char cmdchar, char *newval)
 	tries = 0;
 	while (tries < 6) {
 		upssendchar(NEXT_VAL);
-		upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+		ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+		if ((ret < 1) || (!strcmp(temp, "NO")))
+			return;
 
-		/* TODO: test for "OK", bail out on "NO", etc */
 		upssendchar(cmdchar);
-		upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+		ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+		if ((ret < 1) || (!strcmp(temp, "NO")))
+			return;
 
 		if (!strcmp(temp, newval)) 		/* got it */
 			return;
@@ -477,13 +496,16 @@ void hw_set (char cmdchar, char *newval)
 }
 
 /* set string values in the UPS EEPROM (battery date, local identity) */
-void setstring (char *data, char reqchar)
+static void setstring(const char *data, char reqchar)
 {
 	char	temp[256];
-	int	i;
+	int	i, ret;
 
 	upssendchar(reqchar);
-	upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+	if ((ret < 1) || (!strcmp(temp, "NO")))
+		return;
 
 	upssendchar(NEXT_VAL);
 	usleep(50000);
@@ -500,60 +522,76 @@ void setstring (char *data, char reqchar)
 	}
 
 	usleep(50000);
-	upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
-	usleep(50000);
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
 
-	/* TODO: check for "OK" or "NO" or similar */
+	if ((ret < 1)  || (!strcmp(temp, "NO")))
+		return;
+
+	usleep(50000);
 }
 
-void setvar (int auxcmd, int dlen, char *data)
+int setvar(const char *varname, const char *val)
 {
-	/* TODO: look this up from a table? */
-	switch (auxcmd) {
-		case INFO_WAKEDELAY:
-			hw_set(REQ_WAKEDELAY, data);
-			installinfo(INFO_WAKEDELAY, REQ_WAKEDELAY, ENDCHAR, IGNCHARS);
-			break;
-		case INFO_LOWXFER:
-			hw_set(REQ_LOWXFER, data);
-			installinfo(INFO_LOWXFER, REQ_LOWXFER, ENDCHAR, IGNCHARS);
-			break;
-		case INFO_HIGHXFER:
-			hw_set(REQ_HIGHXFER, data);
-			installinfo(INFO_HIGHXFER, REQ_HIGHXFER, ENDCHAR, IGNCHARS);
-			break;
-		case INFO_LINESENS:
-			hw_set(REQ_LINESENS, data);
-			installinfo(INFO_LINESENS, REQ_LINESENS, ENDCHAR, IGNCHARS);
-			break;
-		case INFO_GRACEDELAY:
-			hw_set(REQ_GRACEDELAY, data);
-			installinfo(INFO_GRACEDELAY, REQ_GRACEDELAY, ENDCHAR, IGNCHARS);
-			break;
-		case INFO_RTHRESH:
-			hw_set(REQ_RTHRESH, data);
-			installinfo(INFO_RTHRESH, REQ_RTHRESH, ENDCHAR, IGNCHARS);
-			break;
-		case INFO_ALRMDEL:
-			hw_set(REQ_ALRMDEL, data);
-			installinfo(INFO_ALRMDEL, REQ_ALRMDEL, ENDCHAR, IGNCHARS);
-			break;
-		case INFO_UPSIDENT:
-			setstring(data, REQ_UPSIDENT);
-			installinfo(INFO_UPSIDENT, REQ_UPSIDENT, ENDCHAR, IGNCHARS);
-			break;
-		case INFO_BATTDATE:
-			setstring(data, REQ_BATTDATE);
-			installinfo(INFO_BATTDATE, REQ_BATTDATE, ENDCHAR, IGNCHARS);
-			break;
-		default:
-			upslogx(LOG_ERR, "setvar: unknown type 0x%04x\n",
-			        auxcmd);
+	if (!strcasecmp(varname, "ups.delay.start")) {
+		hw_set(REQ_WAKEDELAY, val);
+		poll_data("ups.delay.start", REQ_WAKEDELAY);
+		return STAT_SET_HANDLED;
 	}
+
+	if (!strcasecmp(varname, "input.transfer.low")) {
+		hw_set(REQ_LOWXFER, val);
+		poll_data("input.transfer.low", REQ_LOWXFER);
+		return STAT_SET_HANDLED;
+	}
+
+	if (!strcasecmp(varname, "input.transfer.high")) {
+		hw_set(REQ_HIGHXFER, val);
+		poll_data("input.transfer.high", REQ_HIGHXFER);
+		return STAT_SET_HANDLED;
+	}
+
+	if (!strcasecmp(varname, "input.sensitivity")) {
+		hw_set(REQ_LINESENS, val);
+		poll_data("input.sensitivity", REQ_LINESENS);
+		return STAT_SET_HANDLED;
+	}
+
+	if (!strcasecmp(varname, "ups.delay.shutdown")) {
+		hw_set(REQ_GRACEDELAY, val);
+		poll_data("ups.delay.shutdown", REQ_GRACEDELAY);
+		return STAT_SET_HANDLED;
+	}
+
+	if (!strcasecmp(varname, "battery.charge.restart")) {
+		hw_set(REQ_RTHRESH, val);
+		poll_data("battery.charge.restart", REQ_RTHRESH);
+		return STAT_SET_HANDLED;
+	}
+
+	if (!strcasecmp(varname, "battery.alarm.threshold")) {
+		hw_set(REQ_ALRMDEL, val);
+		poll_data("battery.alarm.threshold", REQ_ALRMDEL);
+		return STAT_SET_HANDLED;
+	}
+
+	if (!strcasecmp(varname, "ups.id")) {
+		setstring(val, REQ_UPSIDENT);
+		poll_data("ups.id", REQ_UPSIDENT);
+		return STAT_SET_HANDLED;
+	}
+
+	if (!strcasecmp(varname, "battery.date")) {
+		setstring(val, REQ_BATTDATE);
+		poll_data("battery.date", REQ_BATTDATE);
+		return STAT_SET_HANDLED;
+	}
+
+	upslogx(LOG_NOTICE, "setvar: unknown var [%s]", varname);
+	return STAT_SET_UNKNOWN;
 }
 
 /* check for calibration status and either start or stop */
-void do_cal(int start)
+static void do_cal(int start)
 {
 	char	temp[256];
 	int	tval, ret;
@@ -570,6 +608,9 @@ void do_cal(int start)
 		if (start == 0)	{		/* stop requested */
 			upslogx(LOG_INFO, "Stopping runtime calibration");
 			upssendchar(REQ_CAL);
+
+			/* eat the reply */
+			ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
 			return;
 		}
 
@@ -587,10 +628,13 @@ void do_cal(int start)
 
 	upslogx(LOG_INFO, "Starting runtime calibration");
 	upssendchar(REQ_CAL);
+
+	/* eat the reply */
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
 }
 
-/* handle the CMD_OFF with some paranoia */
-void do_off (void)
+/* handle the off command with some paranoia */
+static void do_off(void)
 {
 	static	time_t lastcmd = 0;
 	time_t	now, elapsed;
@@ -617,43 +661,53 @@ void do_off (void)
 	upssendchar(REQ_OFF);
 }
 
-void instcmd(int auxcmd, int dlen, char *data)
+int instcmd(const char *cmdname, const char *extra)
 {
 	char	temp[256];
 
-	/* TODO: reply to upsd? */
-
-	switch(auxcmd) {
-		case CMD_FPTEST:	/* front panel test */
-			upssendchar(REQ_FPTEST);
-			upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
-			break;
-		case CMD_BTEST1:	/* start battery test */
-			upssendchar(REQ_BTEST);
-			break;
-		case CMD_OFF:		/* power off the load */
-			do_off();
-			break;
-		case CMD_ON:		/* power on the load */
-			upssendchar(REQ_ON);
-			usleep(500000);
-			upssendchar(REQ_ON);
-			break;
-		case CMD_CAL0:		/* stop runtime calibration */
-			do_cal(0);
-			break;
-		case CMD_CAL1:		/* start runtime calibration */
-			do_cal(1);
-			break;
-		default:
-			upslogx(LOG_INFO, "instcmd: unknown type 0x%04x\n",
-			        auxcmd);
+	if (!strcasecmp(cmdname, "test.panel.start")) {
+		upssendchar(REQ_FPTEST);
+		upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+		return STAT_INSTCMD_HANDLED;
 	}
+
+	if (!strcasecmp(cmdname, "test.battery.start")) {
+		upssendchar(REQ_BTEST);
+		return STAT_INSTCMD_HANDLED;
+	}
+
+	if (!strcasecmp(cmdname, "load.off")) {
+		do_off();
+		return STAT_INSTCMD_HANDLED;
+	}
+
+	if (!strcasecmp(cmdname, "load.on")) {
+		upssendchar(REQ_ON);
+		usleep(500000);
+		upssendchar(REQ_ON);
+		return STAT_INSTCMD_HANDLED;
+	}
+
+	if (!strcasecmp(cmdname, "calibrate.stop")) {
+		do_cal(0);
+		return STAT_INSTCMD_HANDLED;
+	}
+
+	if (!strcasecmp(cmdname, "calibrate.start")) {
+		do_cal(1);
+		return STAT_INSTCMD_HANDLED;
+	}
+
+	upslogx(LOG_NOTICE, "instcmd: unknown command [%s]", cmdname);
+	return STAT_INSTCMD_UNKNOWN;
 }
 
 void upsdrv_banner(void)
 {
-	printf ("Network UPS Tools - APC Smart protocol driver 0.60 (%s)\n", UPS_VERSION);
+	printf("Network UPS Tools - APC Smart protocol driver %s (%s)\n", 
+		DRV_VERSION, UPS_VERSION);
+	printf("\n");
+	printf("*** This driver will be replaced by newapc in the future\n");
 }
 
 /* prep the serial port */
@@ -677,30 +731,33 @@ void upsdrv_initinfo(void)
 		printf("Unable to detect an APC Smart protocol UPS on port %s\n", 
 			device_path);
 		printf("Check the cabling, port name or model name and try again\n");
-		exit (1);
+		exit(1);
 	}
 
 	getbaseinfo(device_name);
 
 	/* hardcode the manufacturer - it's pretty obvious here */
-        addinfo (INFO_MFR, "APC", 0, 0);
+        dstate_setinfo("ups.mfr", "APC");
 
-	printf("Detected %s [%s] on %s (level %i)\n", getdata(INFO_MODEL), 
-		getdata(INFO_SERIAL), device_path, upslevel);
+	dstate_setinfo("driver.version.internal", "%s", DRV_VERSION);
+
+	printf("Detected %s [%s] on %s (level %i)\n", 
+		dstate_getinfo("ups.model"), dstate_getinfo("ups.serial"), 
+		device_path, upslevel);
 
 	if (cap_measureupsii)
 		printf("Measure-UPS II detected!\n");
 
 	/* install handlers */
-	upsh.instcmd = instcmd;
-	upsh.setvar = setvar;
+	upsh.new_instcmd = instcmd;
+	upsh.new_setvar = setvar;
 }
 
 /* power down the attached load immediately */
 void upsdrv_shutdown(void)
 {
 	char	temp[32];
-	int	tval;
+	int	tval, ret;
 
 	if (smartmode() == 0)
 		upslogx(LOG_WARNING, "Warning: unable to contact the UPS, but trying the shutdown anyway.");
@@ -708,8 +765,17 @@ void upsdrv_shutdown(void)
 	/* check the line status */
 
 	upssendchar(REQ_STATUS);
-	upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
-	tval = strtol(temp, 0, 16);
+	ret = upsrecv(temp, sizeof(temp), ENDCHAR, IGNCHARS);
+
+	if ((ret < 1) || (!strcmp(temp, "NA"))) {
+		printf("Warning: status unavailable - assuming on battery\n");
+		tval = 64 | 16;		/* LB + OB */
+	} else {
+		tval = strtol(temp, 0, 16);
+	}
+
+	if (testvar("sdtype"))
+		shutdowncmdset = atoi(getval("sdtype"));
 
 	switch (shutdowncmdset) {
 		case 3:		/* shutdown with grace period */
@@ -775,9 +841,7 @@ void upsdrv_makevartable(void)
 	addvar(VAR_VALUE, "sdtype", "Specify shutdown type (1-3)");
 }
 
-/* tell main how many entries we need */
-int upsdrv_infomax(void)
+void upsdrv_cleanup(void)
 {
-	return 128;
 }
 
