@@ -48,7 +48,7 @@
 #define NEXT_VAL	'-'
 #define REQ_FPTEST	'A'
 #define REQ_BTEST	'U'
-#define REQ_SOFTDOWN	'S'
+/* #define REQ_SOFTDOWN	'S' 	not used */
 #define REQ_SHUTDOWN	'K'
 #define REQ_OFF		'Z'
 #define REQ_ON		0x0E	/* ^N */
