@@ -26,6 +26,8 @@
 #include "main.h"
 #include "belkin.h"
 
+#define DRV_VERSION "0.11"
+
 int init_communication(void)
 {
 	int	i;
@@ -51,35 +53,41 @@ void do_status(void)
 
 	send_belkin_command(STATUS,STAT_STATUS,"");
 	res = get_belkin_reply(temp);
-	if (res == -1) 
+	if (res == -1) {
+		dstate_datastale();
 		return;
+	}
 
 	status_init();
 
-	get_belkin_field(temp, st, 6);
-	if (*st == '1')
+	get_belkin_field(temp, st, sizeof(st), 6);
+	if (*st == '1') {
 		status_set("OFF");
 
-	get_belkin_field(temp, st, 2);
-	if (*st == '1') {
-		status_set("OB");
+	} else {	/* (OFF) and (OB | OL) are mutually exclusive */
 
-		send_belkin_command(STATUS,STAT_BATTERY,"");
-		res = get_belkin_reply(temp);
-		if (res == -1)
-			return;
+		get_belkin_field(temp, st, sizeof(st), 2);
+		if (*st == '1') {
+			status_set("OB");
 
-		get_belkin_field(temp, st, 10);
-		res = atoi(st);
-		get_belkin_field(temp, st, 2);
+			send_belkin_command(STATUS,STAT_BATTERY,"");
+			res = get_belkin_reply(temp);
+			if (res == -1)
+				return;
 
-		if (*st == '1' || res < LOW_BAT) 
-			status_set("LB");		/* low battery */
+			get_belkin_field(temp, st, sizeof(st), 10);
+			res = atoi(st);
+			get_belkin_field(temp, st, sizeof(st), 2);
+
+			if (*st == '1' || res < LOW_BAT) 
+				status_set("LB");		/* low battery */
+		}
+		else
+			status_set("OL");	/* on line */
 	}
-	else
-		status_set("OL");	/* on line */
 
 	status_commit();
+	dstate_dataok();
 }
 
 int init_ups_data(void)
@@ -93,43 +101,31 @@ int init_ups_data(void)
 	if (res == -1)
 		return res;
 
-	addinfo(INFO_MODEL, temp, 0, 0);
+	dstate_setinfo("ups.model", "%s", temp);
 
 	send_belkin_command(STATUS, VERSION, "");
 	res = get_belkin_reply(temp);
 	if (res == -1)
 		return (res);
 
-	addinfo(INFO_FIRMREV, temp, 0, 0);
+	dstate_setinfo("ups.firmware", "%s", temp);
 
 	send_belkin_command(STATUS, RATING, "");
 	res = get_belkin_reply(temp);
 
-	get_belkin_field(temp, st, 8);
+	get_belkin_field(temp, st, sizeof(st), 8);
 	low = atof(st) / 0.88;
 
-	get_belkin_field(temp, st, 9);
+	get_belkin_field(temp, st, sizeof(st), 9);
 	high = atof(st) * 0.88;
 
 	tcflush(upsfd,TCIOFLUSH);
 
-	addinfo(INFO_STATUS, "", 0, 0);
-	addinfo(INFO_UTILITY, "", 0, 0);
-	addinfo(INFO_BATTPCT, "", 0, 0);
-	addinfo(INFO_ACFREQ, "", 0, 0);
-	addinfo(INFO_UPSTEMP, "", 0, 0);
-	addinfo(INFO_LOADPCT, "", 0, 0);
-	addinfo(INFO_BATTVOLT, "", 0, 0);
-	addinfo(INFO_OUTVOLT, "", 0, 0);
+	dstate_setinfo("input.transfer.low", "%03.1f", low);
+	dstate_setinfo("input.transfer.high", "%03.1f", high);
 
-	snprintf(st, sizeof(st), "%03.1f", low);
-	addinfo(INFO_LOWXFER, st, 0, 0);
-
-	snprintf(st, sizeof(st), "%03.1f", high);
-	addinfo(INFO_HIGHXFER, st, 0, 0);
-
-	addinfo(INFO_INSTCMD, "", 0, CMD_OFF);
-	addinfo(INFO_INSTCMD, "", 0, CMD_ON);
+	dstate_addcmd("load.off");
+	dstate_addcmd("load.on");
 	upsdrv_updateinfo();
 	return 0;
 }
@@ -148,49 +144,47 @@ void upsdrv_updateinfo(void)
 	if (res == -1)
 		return;
 
-	get_belkin_field(temp, st, 3);
+	get_belkin_field(temp, st, sizeof(st), 3);
 	val = atof(st) / 10;
-	setinfo(INFO_UTILITY, "%05.1f", val);
+	dstate_setinfo("input.voltage", "%05.1f", val);
 
 	send_belkin_command(STATUS,STAT_BATTERY, "");
 	res = get_belkin_reply(temp);
 	if (res == -1)
 		return;
 
-	get_belkin_field(temp, st, 10);
+	get_belkin_field(temp, st, sizeof(st), 10);
 	val = atof(st);
-	setinfo(INFO_BATTPCT, "%03.0f", val);
+	dstate_setinfo("battery.charge", "%03.0f", val);
 
-	get_belkin_field(temp, st, 7);
+	get_belkin_field(temp, st, sizeof(st), 7);
 	val = atof(st) / 10;
-	setinfo(INFO_BATTVOLT, "%4.1f", val);
+	dstate_setinfo("battery.voltage", "%4.1f", val);
 
-	get_belkin_field(temp, st, 9);
+	get_belkin_field(temp, st, sizeof(st), 9);
 	val = atof(st);
-	setinfo(INFO_UPSTEMP, "%03.0f", val);
+	dstate_setinfo("ups.temperature", "%03.0f", val);
 
 	send_belkin_command(STATUS, STAT_OUTPUT, "");
 	res = get_belkin_reply(temp);
 	if (res == -1)
 		return;
 
-	get_belkin_field(temp, st, 2);
+	get_belkin_field(temp, st, sizeof(st), 2);
 	val = atof(st) / 10;
-	setinfo(INFO_ACFREQ, "%.1f", val);
+	dstate_setinfo("input.frequency", "%.1f", val);
 
-	get_belkin_field(temp, st, 4);
+	get_belkin_field(temp, st, sizeof(st), 4);
 	val = atof(st) / 10;
-	setinfo(INFO_OUTVOLT, "%05.1f", val);
+	dstate_setinfo("output.voltage", "%05.1f", val);
 
-	get_belkin_field(temp, st, 7);
+	get_belkin_field(temp, st, sizeof(st), 7);
 	val = atof(st);
-	setinfo(INFO_LOADPCT, "%03.0f", val);
-
-	writeinfo();
+	dstate_setinfo("ups.load", "%03.0f", val);
 }
 
 
-void get_belkin_field(char *temp, char *data, int n)
+void get_belkin_field(char *temp, char *data, size_t datalen, int n)
 {
 	char	st[SMALLBUF];
 	int	i = 0, f = 1;
@@ -212,7 +206,7 @@ void get_belkin_field(char *temp, char *data, int n)
 
 	st[i] = 0;
 
-	strcpy(data, st+f);
+	snprintf(data, datalen, "%s", st+f);
 }
 
 int get_belkin_reply(char *buf)
@@ -252,7 +246,7 @@ void upsdrv_shutdown(void)
 
 	send_belkin_command(STATUS, STAT_STATUS, "");
 	res = get_belkin_reply(temp);
-	get_belkin_field(temp, st, 2);
+	get_belkin_field(temp, st, sizeof(st), 2);
 
         if (*st == '1') {
 		printf("On battery - sending shutdown comand...\n");
@@ -269,7 +263,7 @@ void upsdrv_shutdown(void)
 	send_belkin_command(CONTROL, POWER_ON, "1;15");
 }
 
-/* handle the CMD_OFF with some paranoia */
+/* handle the "load.off" with some paranoia */
 void do_off(void)
 {
 	static	time_t lastcmd = 0;
@@ -296,25 +290,26 @@ void do_off(void)
         send_belkin_command(CONTROL,POWER_OFF,"1;1");
 }
 
-void instcmd(int auxcmd, int dlen, char *data)
+int instcmd(const char *cmdname, const char *extra)
 {
-	switch(auxcmd) {
-		case CMD_OFF:		/* power off the load */
-			do_off();
-			break;
-		case CMD_ON:
-		        send_belkin_command(CONTROL,POWER_ON,"1;1");
-			break;
-		default:
-			upslogx(LOG_INFO, "instcmd: unknown type 0x%04x\n",
-				auxcmd);
+	if (!strcasecmp(cmdname, "load.off")) {
+		do_off();
+		return STAT_INSTCMD_HANDLED;
 	}
+
+	if (!strcasecmp(cmdname, "load.on")) {
+	        send_belkin_command(CONTROL,POWER_ON,"1;1");
+		return STAT_INSTCMD_HANDLED;
+	}
+
+	upslogx(LOG_NOTICE, "instcmd: unknown command [%s]", cmdname);
+	return STAT_INSTCMD_UNKNOWN;
 }
 
 /* install pointers to functions for msg handlers called from msgparse */
 void setuphandlers(void)
 {
-	upsh.instcmd = instcmd;
+	upsh.new_instcmd = instcmd;
 }
 
 int send_string(char *data)
@@ -371,7 +366,8 @@ void set_serialDTR0RTS1(void)
 
 void upsdrv_banner(void)
 {
-	printf("Network UPS Tools - Belkin Smart protocol driver 0.10 (%s)\n", UPS_VERSION);
+	printf("Network UPS Tools - Belkin Smart protocol driver %s (%s)\n",
+		DRV_VERSION, UPS_VERSION);
 }
 
 void upsdrv_help(void)
@@ -406,19 +402,19 @@ void upsdrv_initinfo(void)
 	}
 
 	/* manufacturer ID - hardcoded in this particular module */
-	addinfo(INFO_MFR, "BELKIN", 0, 0);
+	dstate_setinfo("ups.mfr", "BELKIN", 0, 0);
+
+	dstate_setinfo("driver.version.internal", "%s", DRV_VERSION);
 
 	/* see what's out there */
 	init_ups_data();
 
-	printf("Detected %s on %s\n", getdata(INFO_MODEL), 
+	printf("Detected %s on %s\n", dstate_getinfo("ups.model"),
 		device_path);
 
 	setuphandlers();
 }
 
-/* tell main how many entries we need */
-int upsdrv_infomax(void)
+void upsdrv_cleanup(void)
 {
-	return 45;
 }
