@@ -29,27 +29,23 @@
 #define ENDCHAR		'\r'
 #define IGNCHARS	" \n"
 
+#if defined(__sgi) && ! defined(__GNUC__)
+#define        inline  __inline
+#endif
+
 void instcmd (int auxcmd, int dlen, char *data);
 void upsdrv_setvar (int, int, char *);
 
 void upsdrv_initinfo(void)
 {
-	addinfo (INFO_MFR, "Best Power", 0, 0);
-	addinfo (INFO_MODEL, "Fortress", 0, 0);
-	addinfo (INFO_STATUS, "", 0, 0);
-	addinfo (INFO_UTILITY, "", 0, 0);
-	addinfo (INFO_OUTVOLT, "", 0, 0);
-	addinfo (INFO_BATT_RUNTIME, "", 0, 0);
-	addinfo (INFO_BATTVOLT, "", 0, 0);
-	addinfo (INFO_NOMBATVLT, "24", 0, 0);
-	addinfo (INFO_UPSTEMP, "", 0, 0);
-	addinfo (INFO_CURRENT, "", 0, 0);
-	addinfo (INFO_OUT_VA, "", 0, 0);
-	addinfo (INFO_ACFREQ, "", 0, 0);
+	dstate_setinfo("ups.mfr", "Best Power");
+	dstate_setinfo("ups.model", "Fortress");
+	dstate_setinfo("battery.voltage.nominal", "24");
+
 	/*addinfo (INFO_ALRM_OVERLOAD, "", 0, 0);*/ /* Flag */
 	/*addinfo (INFO_ALRM_TEMP, "", 0, 0);*/ /* Flag */
-	
-	addinfo (INFO_PDNGRACE, "10", FLAG_WONLY, 4);
+
+	dstate_setinfo("ups.delay.shutdown", "10");	/* write only */	
 	
 	/* tunable via front panel: (european voltage level)
 	   parameter		factory default  range
@@ -67,15 +63,23 @@ void upsdrv_initinfo(void)
 	   echo off     e0
 	   echo on      e1
 	*/
-	addinfo (INFO_LOWXFER, "", FLAG_RW|FLAG_STRING, 3);
-	addinfo (INFO_HIGHXFER, "", FLAG_RW|FLAG_STRING, 3);
-	addinfo (INFO_LOBATTIME, "", FLAG_RW|FLAG_STRING, 2);
-	
+	dstate_setinfo("input.transfer.low", "");
+	dstate_setflags("input.transfer.low", ST_FLAG_STRING | ST_FLAG_RW);
+	dstate_setaux("input.transfer.low", 3);
+
+	dstate_setinfo("input.transfer.high", "");
+	dstate_setflags("input.transfer.high", ST_FLAG_STRING | ST_FLAG_RW);
+	dstate_setaux("input.transfer.high", 3);
+
+	dstate_setinfo("battery.runtime.low", "");
+	dstate_setflags("battery.runtime.low", ST_FLAG_STRING | ST_FLAG_RW);
+	dstate_setaux("battery.runtime.low", 3);
+
 	upsh.instcmd = instcmd; 
 	upsh.setvar = upsdrv_setvar;
-	
-	addinfo (INFO_INSTCMD, "", 0, CMD_SHUTDOWN);
-	addinfo (INFO_INSTCMD, "", 0, CMD_OFF);
+
+	dstate_addcmd("shutdown.return");
+	dstate_addcmd("load.off");
 }
 
 /* convert hex digit to int */
@@ -163,7 +167,11 @@ void upsdrv_updateinfo(void)
 	   interpret response as hex string, sum of all bytes must be zero
 	*/
 	checksum_ok = (checksum (p) & 0xff) == 0;
-	if (!checksum_ok) return;
+
+	if (!checksum_ok) {
+		dstate_datastale();
+		return;
+	}
 	
 	setinfo_int (INFO_UTILITY, p+24,4);
 	setinfo_int (INFO_OUTVOLT, p+28,4);
@@ -194,7 +202,7 @@ void upsdrv_updateinfo(void)
 	 *	((flags & TIOCM_CTS) == TIOCM_CTS) ? "OB" : "OL");
 	 */
 
-	writeinfo();
+	dstate_dataok();
 }
 
 
@@ -255,8 +263,13 @@ void upsdrv_setvar (int info, int len, char * data)
 
 void upsdrv_shutdown(void)
 {
-	char * grace = getdata (INFO_PDNGRACE);
-	if (!grace) grace = "1"; /* apparently, OFF0 does not work */
+	const	char	*grace;
+
+	grace = getdata(INFO_PDNGRACE);
+
+	if (!grace)
+		grace = "1"; /* apparently, OFF0 does not work */
+
 	printf ("shutdown in %s seconds\n", grace);
 	/* make power return when utility power returns */
 	autorestart (1);
@@ -267,7 +280,7 @@ void upsdrv_shutdown(void)
 
 void instcmd (int auxcmd, int dlen, char *data)
 {
-	char *p;
+	const char *p;
 	
 	switch (auxcmd) {
 	case CMD_OFF:
@@ -341,12 +354,6 @@ void upsdrv_initups(void)
 	 */
 }
 
-/* tell main how many items you need */
-int upsdrv_infomax(void)
+void upsdrv_cleanup(void)
 {
-	/* every addinfo() consumes an entry.  plan ahead, but don't go
-	 * overboard, since this is an easy way to waste memory.
-	 */
-
-	return 128;
 }
