@@ -27,11 +27,18 @@
 
 */
 
+#define DRV_VERSION "0.01"
+
 #include "main.h"
 
 #define ENDCHAR		'\r'
 #define IGNCHARS	"\012"
 
+/* UPS Model Codes */
+#define UNKNOWN         100
+#define ME3100          200
+#define MD1KVA          300 /* Software version P5.05 dated 05/18/89 */
+
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
@@ -58,22 +65,20 @@ struct {
 void  upsdrv_initinfo (void)
 {
 	/* now set up room for all future variables that are supported */
-	addinfo (INFO_MFR, "", 0, 0);
-	addinfo (INFO_MODEL, "", 0, 0);
-	addinfo (INFO_UTILITY, "", 0, 0);
-	addinfo (INFO_OUTVOLT, "", 0, 0);
-	addinfo (INFO_ACFREQ, "", 0, 0);
-	addinfo (INFO_LOADPCT, "", 0, 0);
-	addinfo (INFO_BATTPCT, "", 0, 0);
-	addinfo( INFO_BATTVOLT, "", 0, 0);
-	addinfo (INFO_UPSTEMP, "", 0, 0);
-	addinfo (INFO_STATUS, "", 0, 0);
-	addinfo( INFO_RUNTIME, "", 0, 0);
-  
-	setinfo(INFO_MFR, "%s", "Best Power");
-	setinfo(INFO_MODEL, "Micro Ferrups (ME) %d", fc.va);
-  
-	fprintf(stderr, "Best Power %s detected\n", getdata(INFO_MODEL));
+
+	dstate_setinfo("ups.mfr", "%s", "Best Power");
+	switch(fc.model) {
+          case ME3100:
+	    dstate_setinfo("ups.model", "Micro Ferrups (ME) %d", fc.va);
+	    break;
+          case MD1KVA:
+	    dstate_setinfo("ups.model", "Micro Ferrups (MD) %d", fc.va);
+	    break;
+          default:
+	    exit(1); /* Will never get here, upsdrv_initups() will catch */
+        } 
+	fprintf(stderr, "Best Power %s detected\n", 
+		dstate_getinfo("ups.model"));
 	fprintf(stderr, "Battery voltages %5.1f nominal, %5.1f full, %5.1f empty\n", 
 	 fc.idealbvolts,
 	 fc.fullvolts,
@@ -117,7 +122,7 @@ void upsdrv_updateinfo(void)
       alstat=0, vaout=0;
     double ampsout=0.0, vbatt=0.0, battpercent=0.0, loadpercent=0.0,
       upstemp=0.0, acfreq=0.0;
-    char tmp[16], statstr[128];
+    char tmp[16];
 
     /* Inverter status.  0=off 1=on */
     memcpy(tmp, fstring+16, 2);
@@ -185,12 +190,25 @@ void upsdrv_updateinfo(void)
     upstemp = (double)(atoi(tmp));
 
     /* Percent Load */
-    if (execute("d 16\r", fstring, sizeof(fstring)) > 0) {
-      int l;
-      sscanf(fstring, "16 FullLoad%% %d", &l);
-      loadpercent = (double) l;
+    switch(fc.model) {
+      case ME3100:
+        if (execute("d 16\r", fstring, sizeof(fstring)) > 0) {
+          int l;
+          sscanf(fstring, "16 FullLoad%% %d", &l);
+          loadpercent = (double) l;
+        }
+	break;
+      case MD1KVA:
+        if (execute("d 22\r", fstring, sizeof(fstring)) > 0) {
+          int l;
+          sscanf(fstring, "22 FullLoad%% %d", &l);
+          loadpercent = (double) l;
+        }
+	break;
+      default: /* Will never happen, caught in upsdrv_initups() */
+        fprintf(stderr, "Uknown model in upsdrv_updateinfo()\n");
+        exit(1);
     }
-
     /* Compute battery percent left based on battery voltages. */
     battpercent = ((vbatt - fc.emptyvolts) 
 		   / (fc.fullvolts - fc.emptyvolts) * 100.0);
@@ -201,23 +219,37 @@ void upsdrv_updateinfo(void)
     
     /* Compute status string */
     {
-      int lowbatt, overload, replacebatt, boosting, trimming;
+	int lowbatt, overload, replacebatt, boosting, trimming;
 
-      lowbatt = alstat & (1<<1);
-      overload = alstat & (1<<6);
-      replacebatt = alstat & (1<<10);
-      boosting = inverter && (linestat & (1<<2)) && (vin < 115);
-      trimming = inverter && (linestat & (1<<2)) && (vin > 115);
+	lowbatt = alstat & (1<<1);
+	overload = alstat & (1<<6);
+	replacebatt = alstat & (1<<10);
+	boosting = inverter && (linestat & (1<<2)) && (vin < 115);
+	trimming = inverter && (linestat & (1<<2)) && (vin > 115);
 
-      strcpy(tmp, inverter ? "OB" : "OL");
-      if (lowbatt) strcat (tmp, " LB");
-      if (trimming) strcat (tmp, " TRIM");
-      if (boosting) strcat (tmp, " BOOST");
-      if (replacebatt) strcat (tmp, " RB");
-      if (overload) strcat (tmp, " OVER");
+	status_init();
+      
+	if (inverter) 
+		status_set("OB");
+	else
+		status_set("OL");
 
-      strlcpy(statstr, tmp, sizeof(statstr));
+	if (lowbatt)
+		status_set("LB");
 
+	if (trimming)
+		status_set("TRIM");
+
+	if (boosting)
+		status_set("BOOST");
+
+	if (replacebatt)
+		status_set("RB");
+
+	if (overload)
+		status_set("OVER");
+
+	status_commit();
     }
 
     if (debugging) {
@@ -227,26 +259,29 @@ void upsdrv_updateinfo(void)
       fprintf(stderr,
 	      "      ampsout %5.1f vbatt %5.1f batpcnt %5.1f loadpcnt %5.1f upstemp %5.1f acfreq %5.2f\n",
 	      ampsout, vbatt, battpercent, loadpercent, upstemp, acfreq);
-      fprintf(stderr,
-	      "      STATUS '%s'\n", statstr);
 
     }
 
     /* Stuff information into info structures */
 
-    setinfo(INFO_STATUS, "%s", statstr);
-    setinfo(INFO_UTILITY, "%05.1f", (double)vin);
-    setinfo(INFO_OUTVOLT, "%05.1f", (double)vout);
-    setinfo(INFO_BATTPCT, "%02.1f", battpercent);
-    setinfo(INFO_LOADPCT, "%02.1f", loadpercent);
-    setinfo(INFO_BATTVOLT, "%02.1f", vbatt);
-    setinfo(INFO_ACFREQ, "%05.2f", (double)acfreq);
-    setinfo(INFO_UPSTEMP, "%05.1f", (double)upstemp);
-    setinfo(INFO_RUNTIME, "%d", btimeleft);
-    writeinfo();
-  
-  }
+    dstate_setinfo("input.voltage", "%05.1f", (double)vin);
+    dstate_setinfo("output.voltage", "%05.1f", (double)vout);
+    dstate_setinfo("battery.charge", "%02.1f", battpercent);
+    dstate_setinfo("ups.load", "%02.1f", loadpercent);
+    dstate_setinfo("battery.voltage", "%02.1f", vbatt);
+    dstate_setinfo("input.frequency", "%05.2f", (double)acfreq);
+    dstate_setinfo("ups.temperature", "%05.1f", (double)upstemp);
+    dstate_setinfo("battery.runtime", "%d", btimeleft);
 
+    dstate_dataok();
+
+  } else {
+
+    dstate_datastale();  
+
+  } /* if (execute("f\r", fstring, sizeof(fstring)) > 0) */
+
+  dstate_datastale();
   return;
 }
 
@@ -294,7 +329,8 @@ void upsdrv_help(void)
 
 void upsdrv_banner(void)
 {
-        printf("Network UPS Tools - Best Ferrups Series ME 0.00 (%s)\n\n", UPS_VERSION);
+        printf("Network UPS Tools - Best Ferrups Series ME %s (%s)\n\n", 
+		DRV_VERSION, UPS_VERSION);
 }
 
 static void sync_serial(void) {
@@ -347,6 +383,7 @@ void upsdrv_initups ()
   setup_serial();
   ups_sync();
 
+  fc.model = UNKNOWN;
   /* Obtain Model */
   if (execute("id\r", fcstring, sizeof(fcstring)) < 0) {
     fprintf(stderr, "Failed execute in ups_ident()\n");
@@ -367,40 +404,56 @@ void upsdrv_initups ()
   memcpy(temp, fcstring+9, 2);
   temp[2] = '\0';
 
-  if (memcmp(temp, "ME", 2))  {
-    fprintf(stderr, "Unknown model %s in ups_ident()\n", temp);
-    exit(1);
+  if (memcmp(temp, "ME", 2) == 0)  {
+    fc.model = ME3100;
+  }
+  if (memcmp(temp, "C1", 2) == 0)  {
+    /* Better way to identify unit is using "d 15\r", which results in
+       "15 M#    MD1KVA", "id\r" yields "Unit ID "C1K03588"" */
+    fc.model = MD1KVA;
   }
  
-  /* hard code this to be an ME3100 for now */
-  fc.va = 3100;
-  fc.watts = 2200;
-
-  /* determine shutdown battery voltage */
-  if (execute("d 29\r", fcstring, sizeof(fcstring)) > 0) {
-    sscanf(fcstring, "29 LowBat   %f", &fc.emptyvolts);
+  switch(fc.model) {
+    case ME3100:
+      fc.va = 3100;
+      fc.watts = 2200;
+      /* determine shutdown battery voltage */
+      if (execute("d 29\r", fcstring, sizeof(fcstring)) > 0) {
+        sscanf(fcstring, "29 LowBat   %f", &fc.emptyvolts);
+      }
+      /* determine fully charged battery voltage */
+      if (execute("d 31\r", fcstring, sizeof(fcstring)) > 0) {
+        sscanf(fcstring, "31 HiBatt   %f", &fc.fullvolts);
+      }
+      fc.fullvolts = 54.20;
+      /* determine "ideal" voltage by a guess */
+      fc.idealbvolts = ((fc.fullvolts - fc.emptyvolts) * 0.7) + fc.emptyvolts;
+      break;
+    case MD1KVA:
+      fc.va = 1100;
+      fc.watts = 770; /* Approximate, based on 0.7 power factor */
+      /* determine shutdown battery voltage */
+      if (execute("d 27\r", fcstring, sizeof(fcstring)) > 0) {
+        sscanf(fcstring, "27 LowBatt  %f", &fc.emptyvolts);
+      }
+      /* determine fully charged battery voltage */
+      if (execute("d 28\r", fcstring, sizeof(fcstring)) > 0) {
+        sscanf(fcstring, "28 Hi Batt  %f", &fc.fullvolts);
+      }
+      fc.fullvolts = 13.70;
+      /* determine "ideal" voltage by a guess */
+      fc.idealbvolts = ((fc.fullvolts - fc.emptyvolts) * 0.7) + fc.emptyvolts;
+      break;
+    default:
+      fprintf(stderr, "Uknown model %s in ups_ident()\n", temp);
+      exit(1);
+      break;
   }
 
-  /* determine fully charged battery voltage */
-  if (execute("d 31\r", fcstring, sizeof(fcstring)) > 0) {
-    sscanf(fcstring, "31 HiBatt   %f", &fc.fullvolts);
-  }
-  fc.fullvolts=54.20;
-
-  /* determine "ideal" voltage by a guess */
-  fc.idealbvolts = ((fc.fullvolts - fc.emptyvolts) * 0.7) + fc.emptyvolts;
-
   fc.valid = 1;
   return;
 }
 
-/* tell main how many items you need */
-int upsdrv_infomax(void)
+void upsdrv_cleanup(void)
 {
-	/* every addinfo() consumes an entry.  plan ahead, but don't go
-	 * overboard, since this is an easy way to waste memory.
-	 */
-
-	return 128;
 }
-
