@@ -17,6 +17,8 @@
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
 
+#define DRV_VERSION "1.10"
+
 #include <sys/ioctl.h>
 #include <sys/termios.h>
 
@@ -25,21 +27,83 @@
 
 	int	upstype = -1;
 
+static void parse_output_signals(const char *value, int *line) 
+{ 
+	/* parse signals the serial port can output */ 
+ 
+	*line = 0; 
+ 
+	if (strstr(value, "DTR")) 
+		if (!strstr(value, "-DTR")) 
+			*line |= TIOCM_DTR; 
+ 
+	if (strstr(value, "RTS")) 
+		if (!strstr(value, "-RTS")) 
+			*line |= TIOCM_RTS; 
+ 
+	if (strstr(value, "ST")) 
+		*line |= TIOCM_ST; 
+} 
+ 
+static void parse_input_signals(const char *value, int *line, int *val) 
+{ 
+	/* parse signals the serial port can input */ 
+ 
+	*line = 0; 
+	*val = 0; 
+ 
+	if (strstr(value, "CTS")) 
+		*line |= TIOCM_CTS; 
+	if (!strstr(value, "-CTS")) 
+		*val |= TIOCM_CTS; 
+ 
+	if (strstr(value, "DCD")) 
+		*line |= TIOCM_CD; 
+	if (!strstr(value, "-DCD")) 
+		*val |= TIOCM_CD; 
+ 
+	if (strstr(value, "RNG")) 
+		*line |= TIOCM_RNG; 
+	if (!strstr(value, "-RNG")) 
+		*val |= TIOCM_RNG; 
+}
+
 void upsdrv_initinfo(void)
 {
 	/* setup the basics */
 
 	if (getval("mfr"))
-		addinfo(INFO_MFR, getval("mfr"), 0, 0);
+		dstate_setinfo("ups.mfr", getval("mfr"));
 	else
-		addinfo(INFO_MFR, upstab[upstype].mfr, 0, 0);
+		dstate_setinfo("ups.mfr", upstab[upstype].mfr);
 
 	if (getval("model"))
-		addinfo(INFO_MODEL, getval("model"), 0, 0);
+		dstate_setinfo("ups.model", getval("model"));
 	else
-		addinfo(INFO_MODEL, upstab[upstype].model, 0, 0);
+		dstate_setinfo("ups.model", upstab[upstype].model);
 
-	addinfo(INFO_STATUS, "", 0, 0);
+	if (getval("serial"))
+		dstate_setinfo("ups.serial", getval("serial"));
+
+	dstate_setinfo("driver.version.internal", "%s", DRV_VERSION);
+
+	/* see if the user wants to override the signal definitions */
+
+	if (getval("CP")) 
+		parse_output_signals(getval("CP"), &upstab[upstype].line_norm); 
+ 
+	if (getval("OL")) 
+		parse_input_signals(getval("OL"), &upstab[upstype].line_ol, 
+			&upstab[upstype].val_ol); 
+ 
+	if (getval("LB")) 
+		parse_input_signals(getval("LB"), &upstab[upstype].line_bl, 
+			&upstab[upstype].val_bl); 
+ 
+	if (getval("SD")) 
+		parse_output_signals(getval("SD"), &upstab[upstype].line_sd); 
+
+	dstate_dataok();
 }
 
 /* normal idle loop - keep up with the current state of the UPS */
@@ -51,6 +115,7 @@ void upsdrv_updateinfo(void)
 
 	if (ret != 0) {
 		upslog(LOG_INFO, "ioctl failed");
+		dstate_datastale();
 		return;
 	}
 
@@ -68,7 +133,7 @@ void upsdrv_updateinfo(void)
 		status_set("OB");	/* on battery */
 
 	status_commit();
-	writeinfo();
+	dstate_dataok();
 }
 
 /* show all possible UPS types */
@@ -136,7 +201,8 @@ void upsdrv_help(void)
 
 void upsdrv_banner(void)
 {
-        printf("Network UPS Tools - Generic UPS driver 0.60 (%s)\n", UPS_VERSION);
+        printf("Network UPS Tools - Generic UPS driver %s (%s)\n", 
+		DRV_VERSION, UPS_VERSION);
 }
 
 void upsdrv_makevartable(void)
@@ -144,6 +210,11 @@ void upsdrv_makevartable(void)
 	addvar(VAR_VALUE, "upstype", "Set UPS type (required)");
 	addvar(VAR_VALUE, "mfr", "Override manufacturer name");
 	addvar(VAR_VALUE, "model", "Override model name");
+	addvar(VAR_VALUE, "serial", "Specify the serial number");
+	addvar(VAR_VALUE, "CP", "Override cable power setting"); 
+	addvar(VAR_VALUE, "OL", "Override on line signal"); 
+	addvar(VAR_VALUE, "LB", "Override low battery signal"); 
+	addvar(VAR_VALUE, "SD", "Override shutdown setting"); 
 }
 
 void upsdrv_initups(void)
@@ -153,7 +224,6 @@ void upsdrv_initups(void)
 	open_serial_simple(device_path, 0, upstab[upstype].line_norm);
 }
 
-int upsdrv_infomax(void)
+void upsdrv_cleanup(void)
 {
-	return 8;
 }
