@@ -42,27 +42,30 @@ void upsdrv_initinfo(void)
 	tmp = getval("mfr");
 
 	if (!tmp)
-		addinfo(INFO_MFR, "Liebert", 0, 0);
+		dstate_setinfo("ups.mfr", "Liebert");
 	else
-		addinfo(INFO_MFR, tmp, 0, 0);
+		dstate_setinfo("ups.mfr", "%s", tmp);
 
 	tmp = getval("model");
 
 	if (!tmp)
-		addinfo(INFO_MODEL, "MultiLink", 0, 0);
+		dstate_setinfo("ups.model", "MultiLink");
 	else
-		addinfo(INFO_MODEL, tmp, 0, 0);
-
-	addinfo(INFO_STATUS, "", 0, 0);
+		dstate_setinfo("ups.model", "%s", tmp);
 }
 
+/* require this many OBs or LBs before actually setting it */
+#define DEBOUNCE 3
+
 /* normal idle loop - keep up with the current state of the UPS */
 void upsdrv_updateinfo(void)
 {
 	int	flags, ob, lb, ret;
 	char	c;
+	static	unsigned int	ob_state = 0, ob_last = 0, ob_ctr = 0;
+	static	unsigned int	lb_state = 0, lb_last = 0, lb_ctr = 0;
 
-	ob = 0;
+	ob = lb = 0;
 
 	/* the UPS connects RX to TX when on battery, so test for loopback */
 
@@ -79,23 +82,77 @@ void upsdrv_updateinfo(void)
 
 	if (ret != 0) {
 		upslog(LOG_INFO, "ioctl failed");
+		dstate_datastale();
 		return;
 	}
 
-	lb = (flags & TIOCM_CD);
+	if (flags & TIOCM_CD)
+		lb = 1;
+
+	/* state machine below to ensure status changes are debounced */
+
+	/* OB/OL state change: reset counter */
+	if (ob_last != ob)
+		ob_ctr = 0;
+	else
+		ob_ctr++;
+
+	upsdebugx(2, "OB: state %d last %d now %d ctr %d",
+		ob_state, ob_last, ob, ob_ctr);
+
+	if (ob_ctr >= DEBOUNCE) {
+
+		if (ob != ob_state) {
+
+			upsdebugx(2, "OB: toggling state");
+
+			if (ob_state == 0)
+				ob_state = 1;
+			else
+				ob_state = 0;
+		}
+	}
+
+	ob_last = ob;
+
+	/* now do it again for LB */
+
+	/* state change: reset counter */
+	if (lb_last != lb)
+		lb_ctr = 0;
+	else
+		lb_ctr++;
+
+	upsdebugx(2, "LB: state %d last %d now %d ctr %d",
+		lb_state, lb_last, lb, lb_ctr);
+
+	if (lb_ctr >= DEBOUNCE) {
+
+		if (lb != lb_state) {
+
+			upsdebugx(2, "LB: toggling state");
+
+			if (lb_state == 0)
+				lb_state = 1;
+			else
+				lb_state = 0;
+		}
+	}
+
+	lb_last = lb;
 
 	status_init();
 
-	if (ob)
+	if (ob_state == 1)
 		status_set("OB");	/* on battery */
 	else
 		status_set("OL");	/* on line */
 
-	if (lb)
+	if (lb_state == 1)
 		status_set("LB");	/* low battery */
 
 	status_commit();
-	writeinfo();
+	dstate_dataok();
 }
 
 /*  Open the serial port.  The upscommon routines aren't used because
@@ -110,7 +167,7 @@ static void open_serial_ml(const char *port)
 	signal(SIGALRM, openfail);
 	alarm(3);
 
-	if ((upsfd = open(port, O_RDWR)) == -1)
+	if ((upsfd = open(port, O_RDWR | O_NONBLOCK)) == -1)
 		fatal("Unable to open %s", port);
 
 	alarm(0);
@@ -125,6 +182,7 @@ static void open_serial_ml(const char *port)
 
 	tcgetattr(upsfd, &tio);
 	cfmakeraw(&tio);
+	tio.c_cflag |= CLOCAL;
 	tio.c_cc[VMIN] = 0;
 	tio.c_cc[VTIME] = 1;
 	tcsetattr(upsfd, TCSANOW, &tio);
@@ -153,7 +211,6 @@ void upsdrv_initups(void)
 	open_serial_ml(device_path);
 }
 
-int upsdrv_infomax(void)
+void upsdrv_cleanup(void)
 {
-	return 8;
 }
