@@ -119,7 +119,7 @@ int sec_upsrecv (char *buf)
     struct 	sigaction sa;
     sigset_t 	sec_sigmask;
 
-    strcpy (recvbuf, "");
+    memset(recvbuf, '\0', sizeof(recvbuf));
  
     sa.sa_handler = timeout;
     sigemptyset (&sec_sigmask);
@@ -361,43 +361,54 @@ void sec_setinfo(int varnum, char *value)
 
 void update_pseudovars( void )
 {
-    char  status[256],*v;
-    int   n;
-    float fsum;
+	const char  *v;
+	int   n;
+	float fsum;
 
-    upsdebugx(1, "Synthesizing INFO_STATUS...");
-    /* the STATUS variable */
-    strcpy(status, "");
-    if (supported(INFO_ALRM_OUTPUTOFF) &&
-	strcmp(getdata(INFO_ALRM_OUTPUTOFF), "Output off") == 0)
-	strcat(status, "OFF "); 
-    else if (supported(INFO_OUT_SOURCE)) {
-	if (strcmp(getdata(INFO_OUT_SOURCE), "Normal") == 0)
-	    strcat(status, "OL ");
-	else if (strcmp(getdata(INFO_OUT_SOURCE), "On Battery") == 0)
-	    strcat(status, "OB ");
-	else if (strcmp(getdata(INFO_OUT_SOURCE), "On Bypass") == 0)
-	    strcat(status, "BY ");
-	else if (strcmp(getdata(INFO_OUT_SOURCE), "Reducing") == 0)
-	    strcat(status, "OL TRIM ");
-	else if (strcmp(getdata(INFO_OUT_SOURCE), "Boosting") == 0)
-	    strcat(status, "OL BOOST ");
-    }
-    if (supported(INFO_ALRM_OVERLOAD) &&
-	strcmp(getdata(INFO_ALRM_OVERLOAD), "UPS Overloaded") == 0)
-	strcat(status, "OVER ");
-    if (supported(INFO_BATT_STATUS) &&
-	strcmp(getdata(INFO_BATT_STATUS), "Battery Low") == 0)
-	strcat(status, "LB ");
-    if (supported(INFO_BATT_COND) &&
-	strcmp(getdata(INFO_BATT_COND), "Replace") == 0)
-	strcat(status, "RB ");
+	upsdebugx(1, "Synthesizing INFO_STATUS...");
+	/* the STATUS variable */
 
-    if (supported(INFO_ALRM_SYSOFF) &&
-	strcmp(getdata(INFO_ALRM_SYSOFF), "System off") == 0)
-	strcat(status, "RB ");
+	status_init();
 
-    setinfo (INFO_STATUS, "%s", status);
+	if (supported(INFO_ALRM_OUTPUTOFF) &&
+		strcmp(getdata(INFO_ALRM_OUTPUTOFF), "Output off") == 0) {
+		status_set("OFF");
+	}
+
+	if (supported(INFO_OUT_SOURCE)) {
+		if (strcmp(getdata(INFO_OUT_SOURCE), "Normal") == 0)
+			status_set("OL");
+
+		if (strcmp(getdata(INFO_OUT_SOURCE), "On Battery") == 0)
+			status_set("OB");
+
+		if (strcmp(getdata(INFO_OUT_SOURCE), "On Bypass") == 0)
+			status_set("BYPASS");
+
+		if (strcmp(getdata(INFO_OUT_SOURCE), "Reducing") == 0)
+			status_set("TRIM");
+
+		if (strcmp(getdata(INFO_OUT_SOURCE), "Boosting") == 0)
+			status_set("BOOST");
+	}
+
+	if (supported(INFO_ALRM_OVERLOAD) &&
+		strcmp(getdata(INFO_ALRM_OVERLOAD), "UPS Overloaded") == 0)
+		status_set("OVER");
+
+	if (supported(INFO_BATT_STATUS) &&
+		strcmp(getdata(INFO_BATT_STATUS), "Battery Low") == 0)
+		status_set("LB");
+
+	if (supported(INFO_BATT_COND) &&
+		strcmp(getdata(INFO_BATT_COND), "Replace") == 0)
+		status_set("RB");
+
+	if (supported(INFO_ALRM_SYSOFF) &&
+		strcmp(getdata(INFO_ALRM_SYSOFF), "System off") == 0)
+		status_set("RB");
+
+	status_commit();
 
     upsdebugx(1, "Synthesizing averages...");
     /* Average stats */
@@ -451,11 +462,6 @@ void update_pseudovars( void )
  * Here are the mandatory functions called from the shared main.c
  */
 
-int upsdrv_infomax(void)
-{
-    return 160;
-}
-
 void upsdrv_initinfo(void)
 {
     int msglen, e, v;
@@ -496,10 +502,7 @@ void upsdrv_initinfo(void)
 	    /* add entries for enumerated variable */
 	    while (sec_enumdata[e].type == sec_varlist[v].infotag) {
 		upsdebugx(1, " adding enumval \"%s\" (%d)",sec_enumdata[e].value,e);
-		addinfo(INFO_ENUM,
-			sec_enumdata[e].value,
-			0,
-			sec_enumdata[e].type);
+		addenum(sec_enumdata[e].type, sec_enumdata[e].value);
 		e++;
 	    }
 	}
@@ -569,7 +572,10 @@ void upsdrv_updateinfo(void)
 	    if (sqv(q,f) > 0) {
 		/* only update the value if it's changed */
 		if (strcmp(sec_varlist[sqv(q,f)].value, r) != 0) {
-		    strcpy(sec_varlist[sqv(q,f)].value, r);
+
+		    snprintf(sec_varlist[sqv(q,f)].value, 
+			sizeof(sec_varlist[sqv(q,f)].value), "%s", r);
+
 		    sec_setinfo(sqv(q,f), r);
 		}
 	    }
@@ -581,9 +587,6 @@ void upsdrv_updateinfo(void)
     /* update the non-sec variables */
     update_pseudovars();
     
-    upsdebugx(1, "**Writing updated info");
-    writeinfo();
-
 }
 
 void upsdrv_shutdown(void)
@@ -643,6 +646,9 @@ void upsdrv_banner(void)
 
 void upsdrv_initups(void)
 {
+	/* this driver needs lots of variable conversion work */
+	broken_driver = 1;
+	return;
 
 /*    upssend_delay = 100000; */
 
@@ -664,5 +670,8 @@ void upsdrv_initups(void)
      *	printf("cable is set to %s\n", getval("cable"));
      */
 
-    /* upsh.instcmd = instcmd; */
+}
+
+void upsdrv_cleanup(void)
+{
 }
