@@ -63,3 +63,9 @@ size_t strlcpy(char *dst, const char *src, size_t siz);
 #ifndef CRTSCTS
    #define CRTSCTS   020000000000
 #endif
+
+#ifdef HAVE_GETPASSPHRASE
+#define GETPASS getpassphrase
+#else
+#define GETPASS getpass
+#endif
