@@ -26,25 +26,28 @@
 #endif /* HAVE_SYS_TERMIOS_H */
 
 #include "shared.h"
+#include "upshandler.h"
 
-#ifdef HAVE_MMAP
-#include <sys/types.h>
-#include <sys/mman.h>
-#endif
+/* useful values for new set/instcmd handlers */
+
+#define STAT_INSTCMD_HANDLED	0x0001		/* all we do for now */
+#define STAT_INSTCMD_UNKNOWN	0x0002
+/* FUTURE: add details: pass/fail, etc */
+
+#define STAT_SET_HANDLED	0x0010
+#define STAT_SET_UNKNOWN	0x0011
+/* FUTURE: same */
 
 /* called by alarm signal when the port open fails */
 void openfail(int sig);
 
 /* lock the serial port using flock or uu_lock */
-void lockport (int upsfd, const char *port);
+void lockport(int upsfd, const char *port);
 
 /* try to open the port at a given speed */
 void open_serial_simple(const char *port, speed_t speed, int flags);
 void open_serial(const char *port, speed_t speed);
 
-/* put a notice in the syslog */
-void notice (const char *msg);
-
 /* function for erasing "timeout"-conditions */
 void nolongertimeout(void);
 
@@ -53,62 +56,37 @@ void timeout(int sig);
 
 /* receive up to buflen bytes from the ups into buf until endchar is read */
 /* any characters received that match members of ignchars are discarded */
-int upsrecv (char *buf, int buflen, char endchar, const char *ignchars);
+int upsrecv(char *buf, int buflen, char endchar, const char *ignchars);
 
 /* read buflen chars and store in buf */
-int upsrecvchars (char *buf, int buflen);
+int upsrecvchars(char *buf, int buflen);
 
 /* send a byte to the ups */
-int upssendchar (char data);
+int upssendchar(char data);
 /* send a string to the ups */
 int upssend(const char *fmt, ...);
 
 /* flush any pending input from the ups, with reporting */
-void upsflushin (int expected, int debugit, const char *ignchars);
+void upsflushin(int expected, int debugit, const char *ignchars);
 
-/* install the data that the ups returns for 'reqchar' in the info
-   array position that has type infotype - endchar/ignchars pass to upsrecv */
-void installinfo (int infotype, char reqchar, char endchar, const char *ignchars);
+/* --- old stuff starts here --- */
 
 /* store data into the array */
 void setinfo(int infotype, const char *fmt, ...);
 
 /* set the flags on an existing member of the array */
-void setflag (int infotype, int newflags);
+void setflag(int infotype, int newflags);
 
 /* return the data in the info array with type infotype */
-char *getdata (int infotype);
-
-/* write data to state file - effectively a no-op for shared memory mode */
-void writeinfo(void);
-
-/* create info array of size numinfo from shared memory or normal */
-/* shmok == 1 to allow shared memory, 0 to deny */
-void create_info(int numinfo, int shmok);
-
-/* call this after calling create_info and then doing all your add/setinfos */
-void info_ready(void);
-
-/* create SysV IPC message queue if possible */
-void createmsgq(void);
-
-/* get a message from the queue and parse it - returns 1 if successful */
-int getupsmsg(int wait);
-
-/* place a message into the queue */
-void sendupsmsg (const msgtype *buf);
-
-/* lightweight way to send just an OK/ERR response to a command */
-void msgreply (int reptype);
+const char *getdata(int infotype);
 
 /* add another type to the info array */
-void addinfo (int type, const char *value, int flags, int auxdata);
+void addinfo(int type, const char *value, int flags, int auxdata);
 
 /* add a new ENUM info entry and do other related housekeeping */
-void addenum (int basetype, const char *value);
+void addenum(int basetype, const char *value);
 
-/* close the state fd - used by the exiting parent */
-void close_statefd(void);
+/* --- old stuff ends here --- */
 
 /* clean out the temp space for a new pass */
 void status_init(void);
@@ -116,18 +94,9 @@ void status_init(void);
 /* add a status element */
 void status_set(char *buf);
 
-/* write the status_buf into the info array */
+/* write the temporary status_buf into ups.status */
 void status_commit(void);
 
-/* structure for funcs that get called by msg parse routine */
-struct ups_handler
-{
-	void	(*setvar)(int, int, char *);
-	void	(*instcmd)(int, int, char *);
-};
-
-extern itype *info;
-extern int infomax;
 extern struct ups_handler upsh;
 extern int upsfd;
 extern int flag_timeoutfailure;
@@ -135,37 +104,9 @@ extern unsigned int upssend_delay;
 
 extern int experimental_driver;
 
-#define SIZE_OF_STATEFN 512
-extern char statefn[SIZE_OF_STATEFN];
-
 void rtrim(char *in, char sep);
 void unlockport(int upsfd, const char *port);
+void exit_cleanup(void);
 
-#ifdef HAVE_SHMAT
-
-/*
- * The use of SHM_R and SHM_W is not portable. Neither Posix nor
- * SUSv2 specify it. Solaris and Linux have them (as well as newer
- * versions of NetBSD) but either do not document them (Solaris) or
- * explicitly depreceate their use (NetBSD).
- */
-#ifdef SHM_R
-#define IPC_MODE IPC_CREAT|SHM_R|SHM_W|S_IRGRP|S_IWGRP
-#else
-#define IPC_MODE IPC_CREAT|0600|S_IRGRP|S_IWGRP
-#endif
-
-#endif	/* HAVE_SHMAT */
-
-/* Linux doesn't have MAP_NOSYNC, but it's the default, so... */
-
-#ifndef MAP_NOSYNC
-#define MAP_NOSYNC 0
-#endif
-
-/* Linux libc5 doesn't have MAP_FAILED */
-
-#ifndef MAP_FAILED
-#define MAP_FAILED (void *) -1 
-#endif
-
+/* install handlers for TERM/QUIT/INT and HUP/PIPE */
+void setup_signals(void);
