@@ -20,19 +20,19 @@
 #include "upsd.h"
 #include "conf.h"
 #include "upsconf.h"
-
-#include "parseconf.h"
+#include "sstate.h"
+#include "access.h"
+#include "user.h"
+#include "access.h"
 
 	extern	int	maxage;
-	extern	char	*statepath, *certfile;
+	extern	char	*statepath, *datapath, *certfile;
 	extern	upstype	*firstups;
-	extern	acltype	*firstacl;
-	extern	acctype	*firstacc;
 	ups_t	*upstable = NULL;
 	int	reload_flag;
 
 /* return 1 if usable, 0 if not */
-int parse_upsd_conf_args(int numargs, char **arg)
+static int parse_upsd_conf_args(int numargs, char **arg)
 {
 	/* everything below here uses up through arg[1] */
 	if (numargs < 2)
@@ -53,6 +53,15 @@ int parse_upsd_conf_args(int numargs, char **arg)
 		return 1;
 	}
 
+	/* DATAPATH <dir> */
+	if (!strcmp(arg[0], "DATAPATH")) {
+		if (datapath)
+			free(datapath);
+
+		datapath = xstrdup(arg[1]);
+		return 1;
+	}
+
 	/* CERTFILE <dir> */
 	if (!strcmp(arg[0], "CERTFILE")) {
 		if (certfile)
@@ -68,22 +77,17 @@ int parse_upsd_conf_args(int numargs, char **arg)
 
 	/* ACL <aclname> <ip block> */
 	if (!strcmp(arg[0], "ACL")) {
-		addacl(arg[1], arg[2]);
+		acl_add(arg[1], arg[2]);
 		return 1;
 	}
 
 	if (numargs < 4)
 		return 0;
 
-	/* ACCESS <action> <level> <aclname> [<password>] */
+	/* ACCESS <action> <level> <aclname> */
+	/* any password on the end is now ignored - no more host-based auth */
 	if (!strcmp(arg[0], "ACCESS")) {
-
-		/* don't reference arg[4] unless we have it */
-		if (numargs >= 5)
-			addaccess(arg[1], arg[2], arg[3], arg[4]);
-		else
-			addaccess(arg[1], arg[2], arg[3], NULL);
-
+		access_add(arg[1], arg[2], arg[3]);
 		return 1;
 	}
 
@@ -91,52 +95,63 @@ int parse_upsd_conf_args(int numargs, char **arg)
 	return 0;
 }
 
-/* complain if conf_args can't handle a line */
-void upsd_conf_args(int numargs, char **arg)
+/* called for fatal errors in parseconf like malloc failures */
+void upsd_conf_err(const char *errmsg)
 {
-	if (numargs < 1)
-		return;
-
-	if (!parse_upsd_conf_args(numargs, arg)) {
-		int	i;
-		char	errmsg[SMALLBUF];
-
-		snprintf(errmsg, sizeof(errmsg), "upsd.conf: invalid directive");
-
-		for (i = 0; i < numargs; i++)
-			snprintfcat(errmsg, sizeof(errmsg), " %s", arg[i]);
-
-		upslogx(LOG_WARNING, errmsg);
-	}
-}
-
-void upsd_conf_error(int linenum, char *errtext)
-{
-	if (linenum)
-		upslogx(LOG_ERR, "upsd.conf: %s", errtext);
-	else
-		upslogx(LOG_ERR, "upsd.conf line %d: %s", linenum, errtext);
+	upslogx(LOG_ERR, "Fatal error in parseconf (upsd.conf): %s", errmsg);
 }
 
 void read_upsdconf(int reload)
 {
-	int	ret;
 	char	fn[SMALLBUF];
+	PCONF_CTX	ctx;
 
-	snprintf(fn, sizeof(fn), "%s/upsd.conf", CONFPATH);
+	snprintf(fn, sizeof(fn), "%s/upsd.conf", confpath());
+
+	check_perms(fn);
 
 	reload_flag = reload;
 
-	ret = parseconf(fn, upsd_conf_args, upsd_conf_error);
+	pconf_init(&ctx, upsd_conf_err);
+
+	if (!pconf_file_begin(&ctx, fn)) {
+		pconf_finish(&ctx);
 
-	if (ret != 0) {
 		if (!reload)
-			fatal("Can't open %s", fn);
+			fatalx("%s", ctx.errmsg);
 		else {
-			upslog(LOG_ERR, "Reload failed: can't open %s", fn);
+			upslogx(LOG_ERR, "Reload failed: %s", ctx.errmsg);
 			return;
 		}
 	}
+
+	while (pconf_file_next(&ctx)) {
+		if (pconf_parse_error(&ctx)) {
+			upslogx(LOG_ERR, "Parse error: %s:%d: %s",
+				fn, ctx.linenum, ctx.errmsg);
+			continue;
+		}
+
+		if (ctx.numargs < 1)
+			continue;
+
+		if (!parse_upsd_conf_args(ctx.numargs, ctx.arglist)) {
+			int	i;
+			char	errmsg[SMALLBUF];
+
+			snprintf(errmsg, sizeof(errmsg), 
+				"upsd.conf: invalid directive");
+
+			for (i = 0; i < ctx.numargs; i++)
+				snprintfcat(errmsg, sizeof(errmsg), " %s", 
+					ctx.arglist[i]);
+
+			upslogx(LOG_WARNING, errmsg);
+		}
+
+	}
+
+	pconf_finish(&ctx);		
 }
 
 /* callback during parsing of ups.conf */
@@ -158,6 +173,8 @@ void do_upsconf_args(char *upsname, char *var, char *val)
 				tmp->driver = xstrdup(val);
 			if (!strcmp(var, "port")) 
 				tmp->port = xstrdup(val);
+			if (!strcmp(var, "desc"))
+				tmp->desc = xstrdup(val);
 			return;
 		}
 
@@ -168,12 +185,15 @@ void do_upsconf_args(char *upsname, char *var, char *val)
 	tmp->upsname = xstrdup(upsname);
 	tmp->driver = NULL;
 	tmp->port = NULL;
+	tmp->desc = NULL;
 	tmp->next = NULL;
 
 	if (!strcmp(var, "driver"))
 		tmp->driver = xstrdup(val);
 	if (!strcmp(var, "port"))
 		tmp->port = xstrdup(val);
+	if (!strcmp(var, "desc"))
+		tmp->desc = xstrdup(val);
 
 	if (last)
 		last->next = tmp;
@@ -213,9 +233,9 @@ void upsconf_add(int reloading)
 
 			/* if a UPS exists, update it, else add it as new */
 			if ((reloading) && (findups(tmp->upsname) != NULL))
-				redefine_ups(statefn, tmp->upsname);
+				redefine_ups(statefn, tmp->upsname, tmp->desc);
 			else
-				addups(statefn, tmp->upsname);
+				addups(statefn, tmp->upsname, tmp->desc);
 		}
 
 		/* free tmp's resources */
@@ -224,6 +244,8 @@ void upsconf_add(int reloading)
 			free(tmp->driver);
 		if (tmp->port)
 			free(tmp->port);
+		if (tmp->desc)
+			free(tmp->desc);
 		if (tmp->upsname)
 			free(tmp->upsname);
 		free(tmp);
@@ -258,11 +280,20 @@ void delete_ups(upstype *target)
 			else
 				last->next = ptr->next;
 
+			if (ptr->sock_fd != -1)
+				close(ptr->sock_fd);
+
 			/* release memory */
+			sstate_infofree(ptr);
+			sstate_cmdfree(ptr);
+			pconf_finish(&ptr->sock_ctx);
+
 			if (ptr->fn)
 				free(ptr->fn);
 			if (ptr->name)
 				free(ptr->name);
+			if (ptr->desc)
+				free(ptr->desc);
 			free(ptr);
 
 			return;
@@ -276,53 +307,13 @@ void delete_ups(upstype *target)
 	upslogx(LOG_ERR, "delete_ups: UPS not found");
 }			
 
-void delete_acls(void)
-{
-	acltype	*ptr, *next;
-
-	ptr = firstacl;
-
-	while (ptr) {
-		next = ptr->next;
-		
-		if (ptr->name)
-			free(ptr->name);
-		free(ptr);
-
-		ptr = next;
-	}
-
-	firstacl = NULL;
-}
-
-void delete_access(void)
-{
-	acctype	*ptr, *next;
-
-	ptr = firstacc;
-
-	while (ptr) {
-		next = ptr->next;
-
-		if (ptr->aclname)
-			free(ptr->aclname);
-		if (ptr->password)
-			free(ptr->password);
-		free(ptr);
-
-		ptr = next;
-	}
-
-	firstacc = NULL;
-}	
-
 /* see if we can open a file */
 static int check_file(char *fn)
 {
 	char	chkfn[SMALLBUF];
 	FILE	*f;
 
-	snprintf(chkfn, sizeof(chkfn), "%s/%s", CONFPATH, fn);
+	snprintf(chkfn, sizeof(chkfn), "%s/%s", confpath(), fn);
 
 	f = fopen(chkfn, "r");
 
@@ -358,8 +349,8 @@ void reload_config(void)
 	upsconf_add(1);			/* 1 = reloading */
 
 	/* flush ACL/ACCESS definitions */
-	delete_acls();
-	delete_access();
+	acl_free();
+	access_free();
 
 	/* now reread upsd.conf */
 	read_upsdconf(1);
